from ._base import ValidatorBase
from .. import Invalid
from ..schema.util import get_primitive_name, get_type_name


class Type(ValidatorBase):
    """ Check if the value has the specific type with `isinstance()` check.

    In contrast to [Schema types](#schema) which performs a strict check, this check is relaxed and accepts subtypes
    as well.

    ```python
    from good import Schema, Type

    schema = Schema(Type(int))
    schema(1)  #-> 1
    schema(True)  #-> True
    ```

    :param type: The type to check instances against.
    :type type: type
    """
    def __init__(self, type):
        self.type = type
        self.name = get_type_name(type)

    def __call__(self, v):
        # Type check
        if not isinstance(v, self.type):
            raise Invalid(_(u'Wrong type'), provided=get_type_name(type(v)))
        # Fine
        return v


class Coerce(ValidatorBase):
    """ Coerce a value to a type with the provided callable.

    `Coerce` applies the *constructor* to the input value and returns a value cast to the provided type.

    If *constructor* fails with `TypeError` or `ValueError`, the value is considered invalid and `Coerce` complains
    on that with a custom message.

    However, if *constructor* raises [`Invalid`](#invalid) -- the error object is used as it.

    ```python
    from good import Schema, Coerce

    schema = Schema(Coerce(int))
    schema(u'1')  #-> 1
    schema(u'a')
    #-> Invalid: Invalid value: expected *Integer number, got a
    ```

    :param constructor: Callable that typecasts the input value
    :type constructor: callable|type
    """

    def __init__(self, constructor):
        self.constructor = constructor
        self.name = _(u'*{type}').format(type=get_primitive_name(constructor))

    def __call__(self, v):
        try:
            return self.constructor(v)
        except (TypeError, ValueError):
            raise Invalid(_(u'Invalid value'))

__all__ = ('Type', 'Coerce',)



"""
Source: [good/schema/errors.py](good/schema/errors.py)

When [validating user input](#validating), [`Schema`](#schema) collects all errors and throws these
after the whole input value is validated. This makes sure that you can report *all* errors at once.

With simple schemas, like `Schema(int)`, only a single error is available: e.g. wrong value type.
In this case, [`Invalid`](#invalid) error is raised.

However, with complex schemas with embedded structures and such, multiple errors can occur:
then [`MultipleInvalid`] is reported.

All errors are available right at the top-level:

```python
from good import Invalid, MultipleInvalid
```
"""

import six


class BaseError(Exception):
    """ Base validation exception """


class SchemaError(BaseError):
    """ Schema error (e.g. malformed) """


class Invalid(BaseError):
    """ Validation error for a single value.

    This exception is guaranteed to contain text values which are meaningful for the user.

    :param message: Validation error message.
    :type message: unicode
    :param expected: Expected value: info about the value the validator was expecting.

        If validator does not specify it -- the name of the validator is used.

    :type expected: unicode
    :param provided: Provided value: info about the value that was actually supplied by the user

        If validator does not specify it -- the input value is typecasted to string and stored here.

    :param provided: unicode
    :param path: Path to the error value.

        E.g. if an invalid value was encountered at ['a'].b[1], then path=['a', 'b', 1].

    :type path: list
    :param validator: The validator that has failed: a schema item
    :type validator: *
    :param info: Custom values that might be provided by the validator. No built-in validator uses this.
    :type info: dict
    """

    def __init__(self, message, expected=None, provided=None, path=None, validator=None, **info):
        super(Invalid, self).__init__(message, expected, provided, path, validator)
        self.message = message
        self.expected = expected
        self.provided = provided
        self.path = path or []
        self.validator = validator
        self.info = info

    def __iter__(self):
        """ Iterate over container errors.

        For `Invalid`, just yields self, however for `MultipleInvalid` it yields every contained errors.

        Hence, it allows to iterate all errors without checking whether it's a multi-error or not.
        """
        yield self

    def __repr__(self):
        return '{cls}({0.message!r}, ' \
               'expected={0.expected!r}, ' \
               'provided={0.provided!r}, ' \
               'path={0.path!r}, ' \
               'validator={0.validator!r}, ' \
               'info={0.info!r})' \
            .format(self, cls=type(self).__name__,)

    def __str__(self):
        return six.text_type(self).encode('utf8')

    def __unicode__(self):
        return u'{message}: expected {0.expected}, got {0.provided}'.format(
            self,
            message=self.message if not self.path else u'{} @ {}'.format(
                self.message,
                u''.join(map(
                    lambda v: u'[{!r}]'.format(v),
                    self.path
                ))
            )
        )

    def enrich(self, expected=None, provided=None, path=None, validator=None):
        """ Enrich this error with additional information.

        This works with both Invalid and MultipleInvalid (thanks to `Invalid` being iterable):
        in the latter case, the defaults are applied to all collected errors.

        The specified arguments are only set on `Invalid` errors which do not have any value on the property.

        One exclusion is `path`: if provided, it is prepended to `Invalid.path`.
        This feature is especially useful when validating the whole input with multiple different schemas:

        ```python
        from good import Schema, Invalid

        schema = Schema(int)
        input = {
            'user': {
                'age': 10,
            }
        }

        try:
            schema(input['user']['age'])
        except Invalid as e:
            e.enrich(path=['user', 'age'])  # Make the path reflect the reality
            raise  # re-raise the error with updated fields
        ```

        This is used when validating a value within a container.

        :param expected: Invalid.expected default
        :type expected: unicode|None
        :param provided: Invalid.provided default
        :type provided: unicode|None
        :param path: Prefix to prepend to Invalid.path
        :type path: list|None
        :param validator: Invalid.validator default
        :rtype: Invalid|MultipleInvalid
        """
        for e in self:
            # defaults on fields
            if e.expected is None and expected is not None:
                e.expected = expected
            if e.provided is None and provided is not None:
                e.provided = provided
            if e.validator is None and validator is not None:
                e.validator = validator
            # path prefix
            e.path = (path or []) + e.path
        return self

    if six.PY3:
        __bytes__, __str__ = __str__, __unicode__


class MultipleInvalid(Invalid):
    """ Validation errors for multiple values.

    This error is raised when the [`Schema`](#schema) has reported multiple errors, e.g. for several dictionary keys.

    `MultipleInvalid` has the same attributes as [`Invalid`](#invalid),
    but the values are taken from the first error in the list.

    In addition, it has the `errors` attribute, which is a list of [`Invalid`](#invalid) errors collected by the schema.
    The list is guaranteed to be plain: e.g. there will be no underlying hierarchy of `MultipleInvalid`.

    Note that both `Invalid` and `MultipleInvalid` are iterable, which allows to process them in singularity:

    ```python
    try:
        schema(input_value)
    except Invalid as ee:
        reported_problems = {}
        for e in ee:  # Iterate over `Invalid`
            path_str = u'.'.join(e.path)  # 'a.b.c.d', JavaScript-friendly :)
            reported_problems[path_str] = e.message
        #.. send reported_problems to the user
    ```

    In this example, we create a dictionary of paths (as strings) mapped to error strings for the user.

    :param errors: The reported errors.

        If it contains `MultipleInvalid` errors -- the list is recursively flattened
        so all of them are guaranteed to be instances of [`Invalid`](#invalid).

    :type errors: list[Invalid]
    """

    def __init__(self, errors):
        # Flatten errors
        errors = self.flatten(errors)

        # Create from errors
        e = errors[0]
        super(MultipleInvalid, self).__init__(e.message, e.expected, e.provided, e.path, e.validator, **e.info)

        #: The collected errors
        self.errors = errors

    def __iter__(self):
        return iter(self.errors)

    def __repr__(self):
        return '{cls}({0!r})'.format(self.errors, cls=type(self).__name__)

    @classmethod
    def flatten(cls, errors):
        """ Unwind `MultipleErrors` to have a plain list of `Invalid`

        :type errors: list[Invalid|MultipleInvalid]
        :rtype: list[Invalid]
        """
        ers = []
        for e in errors:
            if isinstance(e, MultipleInvalid):
                ers.extend(cls.flatten(e.errors))
            else:
                ers.append(e)
        return ers

    @classmethod
    def if_multiple(cls, errors):
        """ Provided a list of errors, choose which one to throw: `Invalid` or `MultipleInvalid`.

        `MultipleInvalid` is only used for multiple errors.

        :param errors: The list of collected errors
        :type errors: list[Invalid]
        :rtype: Invalid|MultipleInvalid
        """
        assert errors, 'Errors list is empty'
        return errors[0] if len(errors) == 1 else MultipleInvalid(errors)



import good, good.schema.errors, good.voluptuous
from exdoc import doc, getmembers

import json

doccls = lambda cls, *allowed_keys: {
    'cls': doc(cls),
    'attrs': {name: doc(m) for name, m in getmembers(cls, None, lambda key, value: key in allowed_keys or not key.startswith('_'))}
}

joindict = lambda *dicts: reduce(lambda memo, d: memo.update(d) or memo, dicts, {})

docmodule = lambda mod: {
    'module': doc(mod),
    'members': [ doc(getattr(mod, name)) for name in mod.__all__]
}

data = {
    'good': doc(good),
    'voluptuous': doc(good.voluptuous),

    'Schema': doccls(good.Schema, None, '__call__'),
    'errors': doc(good.schema.errors),
    'Invalid': doccls(good.Invalid),
    'MultipleInvalid': doccls(good.MultipleInvalid),
    'markers': docmodule(good.markers),

    'helpers': docmodule(good.helpers),

    'predicates': docmodule(good.validators.predicates),
    'types': docmodule(good.validators.types),
    'values': docmodule(good.validators.values),
    'boolean': docmodule(good.validators.boolean),
    'numbers': docmodule(good.validators.numbers),
    'strings': docmodule(good.validators.strings),
    'files': docmodule(good.validators.files),
}

# Patches


class MyJsonEncoder(json.JSONEncoder):
    def default(self, o):
        # Classes
        if isinstance(o, type):
            return o.__name__
        return super(MyJsonEncoder, self).default(o)

print json.dumps(data, indent=2, cls=MyJsonEncoder)



import collections

from ._base import ValidatorBase
from .. import Invalid
from ..schema.util import get_literal_name, get_type_name, get_primitive_name, const


class In(ValidatorBase):
    """ Validate that a value is in a collection.

    This is a plain simple `value in container` check, where `container` is a collection of literals.

    In contrast to [`Any`](#any), it does not compile its arguments into schemas,
    and hence achieves better performance.

    ```python
    from good import Schema, In

    schema = Schema(In({1, 2, 3}))

    schema(1)  #-> 1
    schema(99)
    #-> Invalid: Value not allowed: expected In(1,2,3), got 99
    ```

    The same example will work with [`Any`](#any), but slower :-)

    :param container: Collection of allowed values.

        In addition to naive tuple/list/set/dict, this can be any object that supports `in` operation.
    :type container: collections.Container
    """

    def __init__(self, container):
        assert isinstance(container, collections.Container), '`container` must support `in` operation'
        self.container = container

        self.name = _(u'In({container})').format(
            container=
                _(u',').join(map(get_literal_name, self.container))  # iterable
                if isinstance(container, collections.Iterable) else
                get_primitive_name(self.container)  # not iterable
        )

    def __call__(self, v):
        # Test
        if v not in self.container:
            raise Invalid(_(u'Value not allowed'))

        # Okay
        return v


class Length(ValidatorBase):
    """ Validate that the provided collection has length in a certain range.

    ```python
    from good import Schema, Length

    schema = Schema(All(
        # Ensure it's a list (and not any other iterable type)
        list,
        # Validate length
        Length(max=3),
    ))
    ```

    Since mappings also have length, they can be validated as well:

    ```python
    schema = Schema({
        # Strings mapped to integers
        str: int,
        # Size = 1..3
        # Empty dicts are not allowed since `str` is implicitly `Required(str)`
        Entire: Length(max=3)
    })

    schema([1])  #-> ok
    schema([1,2,3,4])
    #-> Invalid: Too many values (3 is the most): expected Length(..3), got 4
    ```

    :param min: Minimal allowed length, or `None` to impose no limits.
    :type min: int|None
    :param max: Maximal allowed length, or `None` to impose no limits.
    :type max: int|None
    """

    def __init__(self, min=None, max=None):
        # `min` validator
        self.min_error = lambda length: Invalid(_(u'Too few values ({min} is the least)').format(min=min),
                                                get_literal_name(min), get_literal_name(length))
        self.min = min

        # `max` validator
        self.max_error = lambda length: Invalid(_(u'Too many values ({max} is the most)').format(max=max),
                                                get_literal_name(max), get_literal_name(length))
        self.max = max

        # Name
        self.name = _(u'Length({min}..{max})').format(
            min=_(u'') if min is None else min,
            max=_(u'') if max is None else max
        )

    def __call__(self, v):
        if not isinstance(v, collections.Sized):
            raise Invalid(_(u'Input is not a collection'), u'Collection', get_type_name(type(v)))

        length = len(v)

        # Validate
        if self.min is not None and length < self.min:
            raise self.min_error(length)
        if self.max is not None and length > self.max:
            raise self.max_error(length)

        # Ok
        return v


class Default(ValidatorBase):
    """ Initialize a value to a default if it's not provided.

    "Not provided" means `None`, so basically it replaces `None`s with the default:

    ```python
    from good import Schema, Any, Default

    schema = Schema(Any(
        # Accept ints
        int,
        # Replace `None` with 0
        Default(0)
    ))

    schema(1)  #-> 1
    schema(None)  #-> 0
    ```

    It raises [`Invalid`](#invalid) on all values except for `None` and `default`:

    ```python
    schema = Schema(Default(42))

    schema(42)  #-> 42
    schema(None)  #-> 42
    schema(1)
    #-> Invalid: Invalid value
    ```

    In addition, `Default` has special behavior with `Required` marker which is built into it:
    if a required key was not provided -- it's created with the default value:

    ```python
    from good import Schema, Default

    schema = Schema({
        # remember that keys are implicitly required
        'name': str,
        'age': Any(int, Default(0))
    })

    schema({'name': 'Alex'})  #-> {'name': 'Alex', 'age': 0}
    ```

    :param default: The default value to use
    """
    def __init__(self, default):
        self.default = default
        self.name = _(u'Default={default}').format(default=default)

    def __call__(self, v):
        if v is None or v is const.UNDEFINED or v == self.default:
            return self.default
        raise Invalid(_(u'Invalid value'))


class Fallback(Default):
    """ Always returns the default value.

    Works like [`Default`](#default), but does not fail on any values.

    Typical usage is to terminate [`Any`](#any) chain in case nothing worked:

    ```python
    from good import Schema, Any, Fallback

    schema = Schema(Any(
        int,
        # All non-integer numbers are replaced with `None`
        Fallback(None)
    ))
    ```

    Like [`Default`](#default), it also works with mappings.

    Internally, `Default` and `Fallback` work by feeding the schema with a special [`Undefined`](good/schema/util.py) value:
    if the schema manages to return some value without errors -- then it has the named "default behavior",
    and this validator just leverages the feature.

    A "fallback value" may be provided manually, and will work absolutely the same
    (since value schema manages to succeed even though `Undefined` was given):

    ```python
    schema = Schema({
        'name': str,
        'age': Any(int, lambda v: 42)
    })
    ```

    :param default: The value that's always returned
    """

    def __call__(self, v):
        return self.default


__all__ = ('In', 'Length', 'Default', 'Fallback')



import six

from .compiler import CompiledSchema
from . import markers


class Schema(object):
    """ Validation schema.

    A schema is a Python structure where nodes are pattern-matched against the corresponding values.
    It leverages the full flexibility of Python, allowing you to match values, types, data structures and much more.

    When a schema is created, it's compiled into a callable function which does the validation, hence it does not need
    to analyze the schema every time.

    Once the Schema is defined, validation can be triggered by calling it:

    ```python
    from good import Schema

    schema = Schema({ 'a': str })
    # Test
    schema({ 'a': 'i am a valid string' })
    ```

    The following rules exist:

    1. **Literal**: plain value is validated with direct comparison (equality check):

        ```python
        Schema(1)(1)  #-> 1
        Schema(1)(2)  #-> Invalid: Invalid value: expected 1, got 2
        ```

    2. **Type**: type schema produces a strict `type(v) == schema` check on the input value:

        ```python
        Schema(int)(1)    #-> 1
        Schema(int)(True)
        #-> Invalid: Wrong type: expected Integer number, got Boolean
        Schema(int)('1')
        #-> Invalid: Wrong type: expected Integer number, got Binary String
        ```

        For Python2, there is an exception for `basestring`: it won't make strict type checks, but rather `isinstance()`.

        For a relaxed `isinstance()` check, see [`Type`](#type) validator.

    3. **Callable**: is applied to the value and the result is used as the final value.

       Callables should raise [`Invalid`](#invalid) errors in case of a failure, however some generic error types are
       converted automatically: see [Callables](#callables).

       In addition, validators are allowed to transform a value to the required form.
       For instance, [`Coerce(int)`](#coerce) returns a callable which will convert input values into `int` or fail.

       ```python
       def CoerceInt(v):  # naive Coerce(int) implementation
           return int(v)

       Schema(CoerceInt)(1)    #-> 1
       Schema(CoerceInt)('1')  #-> 1
       Schema(CoerceInt)('a')
       #-> Invalid: invalid literal for int(): expected CoerceInt(), got a
       ```

    4. **`Schema`**: a schema may contain sub-schemas:

        ```python
        sub_schema = Schema(int)
        schema = Schema([None, sub_schema])

        schema([None, 1, 2])  #-> [None, 1, 2]
        schema([None, '1'])  #-> Invalid: invalid value
        ```

        Since `Schema` is callable, validation transparently by just calling it :)

    Moreover, instances of the following types are converted to callables on the compilation phase:

    1. **Iterables** (`list`, `tuple`, `set`, custom iterables):

        Iterables are treated as a set of valid values,
        where each value in the input is compared against each value in the schema.

        In order for the input to be valid, it needs to have the same iterable type, and all of its
        values should have at least one matching value in the schema.

        ```python
        schema = Schema([1, 2, 3])  # List of valid values

        schema([1, 2, 2])  #-> [1, 2, 2]
        schema([1, 2, 4])  #-> Invalid: Invalid value @ [2]: expected List[1|2|3], got 4
        schema((1, 2, 2))  #-> Invalid: Wrong value type: expected List, got Tuple
        ```

        Each value within the iterable is a schema as well, and validation requires that
        each member of the input value matches *any* of the schemas.
        Thus, an iterable is a way to define *OR* validation rule for every member of the iterable:

        ```python
        Schema([ # All values should be
            # .. int ..
            int,
            # .. or a string, casted to int ..
            lambda v: int(v)
        ])([ 1, 2, '3' ])  #-> [ 1, 2, 3 ]
        ```

        This example works like this:

        1. Validate that the input value has the matching type: `list` in this case
        2. For every member of the list, test that there is a matching value in the schema.

            E.g. for value `1` -- `int` matches (immediate `instanceof()` check).
            However, for value `'3'` -- `int` fails, but the callable manages to do it with no errors,
            and transforms the value as well.

            Since lists are ordered, the first schema that didn't fail is used.

    2. **Mappings** (`dict`, custom mappings):

        Each key-value pair in the input mapping is validated against the corresponding schema pair:

        ```python
        Schema({
            'name': str,
            'age': lambda v: int(v)
        })({
            'name': 'Alex',
            'age': '18',
        })  #-> {'name': 'Alex', 'age': 18}
        ```

        When validating, *both* keys and values are schemas, which allows to use nested schemas and interesting validation rules.
        For instance, let's use [`In`](#in) validator to match certain keys:

        ```python
        from good import Schema, In

        Schema({
            # These two keys should have integer values
            In('age', 'height'): int,
            # All other keys should have string values
            str: str,
        })({
            'age': 18,
            'height': 173,
            'name': 'Alex',
        })
        ```

        This works like this:

        1. Test that the input has a matching type (`dict`)
        2. For each key in the input mapping, matching keys are selected from the schema
        3. Validate input values with the corresponding value in the schema.

        In addition, certain keys can be marked as [`Required`](#required) and [`Optional`](#optional).
        The default behavior is to have all keys required, but this can be changed by providing
        `default_keys=Optional` argument to the Schema.

        Finally, a mapping does not allow any extra keys (keys not defined in the schema). To change this, provide
        `extra_keys=Allow` to the `Schema` constructor.

    These are just the basic rules, and for sure `Schema` can do much more than that!
    Additional logic is implemented through [Markers](#markers) and [Validators](#validation-tools),
    which are described in the following chapters.

    ## Callables

    Finally, here are the things to consider when using custom callables for validation:

    * Throwing errors.

        If the callable throws [`Invalid`](#invalid) exception, it's used as is with all the rich info it provides.
        Schema is smart enough to fill into most of the arguments (see [`Invalid.enrich`](#invalidenrich)),
        so it's enough to use a custom message, and probably, set a human-friendly `expected` field.

        In addition, specific error types are wrapped into `Invalid` automatically: these are
        `AssertionError`, `TypeError`, `ValueError`.
        Schema tries to do its best, but such messages will probably be cryptic for the user.
        Hence, always raise meaningful errors when creating custom validators.
        Still, this opens the possibility to use Python typecasting with validators like `lambda v: int(v)`,
        since most of them are throwing `TypeError` or `ValueError`.

    * Naming.

        If the provided callable does not specify `Invalid.expected` expected value,
        the `__name__` of the callable is be used instead.
        E.g. `def intify(v):pass` becomes `'intify()'` in reported errors.

        If a custom name is desired on the callable -- set the `name` attribute on the callable object.
        This works best with classes, however a function can accept `name` attribute as well.

        For convenience, [`@message`](#message) and [`@name`](#name) decorators can be used on callables
        to specify the name and override the error message used when the validator fails.

    * Signals.

        A callable may decide that the value is soooo invalid that it should be dropped from the sanitized output.
        In this case, the callable should raise `good.schema.signals.RemoveValue`.

        This is used by the `Remove()` marker, but can be leveraged by other callables as well.

    ## Priorities

    Every schema type has a priority ([source](good/schema/util.py)),
    which define the sequence for matching keys in a mapping schema:

    1. Literals have highest priority
    2. Types has lower priorities than literals, hence schemas can define specific rules for individual keys,
        and then declare general rules by type-matching:

        ```python
        Schema({
            'name': str,  # Specific rule with a literal
            str: int,     # General rule with a type
        })
        ```
    3. Callables, iterables, mappings -- have lower priorities.

    In addition, [Markers](#markers) have individual priorities,
    which can be higher that literals ([`Remove()`](#remove) marker) or lower than callables ([`Extra`](#extra) marker).
    """

    compiled_schema_cls = CompiledSchema

    def __init__(self, schema, default_keys=None, extra_keys=None):
        """ Creates a compiled `Schema` object from the given schema definition.

        Under the hood, it uses `SchemaCompiler`: see the [source](good/schema/compiler.py) if interested.

        :param schema: Schema definition
        :type schema: *
        :param default_keys: Default mapping keys behavior:
            a [`Marker`](#markers) class used as a default on mapping keys which are not Marker()ed with anything.

            Defaults to `markers.Required`.

        :type default_keys: type
        :param extra_keys: Default extra keys behavior: sub-schema, or a [`Marker`](#markers) class.

            Defaults to `markers.Reject`

        :type extra_keys: *
        :raises SchemaError: Schema compilation error
        """
        self.compiled = self.compiled_schema_cls(
            schema, [],
            default_keys or markers.Required,
            extra_keys or markers.Reject)
        self.name = self.compiled.name

    def __repr__(self):
        return repr(self.compiled)

    def __unicode__(self):
        return six.text_type(self.compiled)

    if six.PY3:
        __str__ = __unicode__

    def __call__(self, value):
        """ Having a [`Schema`](#schema), user input can be validated by calling the Schema on the input value.

        When called, the Schema will return sanitized value, or raise exceptions.

        :param value: Input value to validate
        :return: Sanitized value
        :raises good.Invalid: Validation error on a single value. See [`Invalid`](#invalid).
        :raises good.MultipleInvalid: Validation error on multiple values. See [`MultipleInvalid`](#multipleinvalid).
        """
        return self.compiled(value)



from ._base import ValidatorBase
from .. import Invalid
from ..schema.util import get_literal_name, get_type_name


class Range(ValidatorBase):
    """ Validate that the value is within the defined range, inclusive.
    Raise [`Invalid`](#invalid) error if not.

    ```python
    from good import Schema, Range

    schema = Schema(Range(1, 10))

    schema(1)  #-> 1
    schema(10)  #-> 10
    schema(15)
    #-> Invalid: Value must be at most 10: expected Range(1..10), got 15
    ```

    If the value cannot be compared to a number -- raises [`Invalid`](#invalid).
    Note that in Python2 almost everything can be compared to a number, including strings, dicts and lists!

    :param min: Minimal allowed value, or `None` to impose no limits.
    :type min: int|float|None
    :param max: Maximal allowed value, or `None` to impose no limits.
    :type max: int|float|None
    """

    def __init__(self, min=None, max=None):
        # `min` validator
        self.min_error = lambda: Invalid(_(u'Value must be at least {min}').format(min=min), get_literal_name(min))
        self.min = min

        # `max` validator
        self.max_error = lambda: Invalid(_(u'Value must be at most {max}').format(max=max),  get_literal_name(max))
        self.max = max

        # Name
        self.name = _(u'Range({min}..{max})').format(
            min=_(u'') if min is None else min,
            max=_(u'') if max is None else max
        )

    def __call__(self, v):
        # Validate
        try:
            if self.min is not None and v < self.min:
                raise self.min_error()
            if self.max is not None and v > self.max:
                raise self.max_error()
        except TypeError:  # cannot compare
            raise Invalid(_(u'Value should be a number'), _(u'Number'), get_type_name(type(v)))

        # Ok
        return v


class Clamp(ValidatorBase):
    """ Clamp a value to the defined range, inclusive.

    ```python
    from good import Schema, Clamp

    schema = Schema(Clamp(1, 10))

    schema(-1)  #-> 1
    schema(1)  #-> 1
    schema(10)  #-> 10
    schema(15)  #-> 10
    ```

    If the value cannot be compared to a number -- raises [`Invalid`](#invalid).
    Note that in Python2 almost everything can be compared to a number, including strings, dicts and lists!

    :param min: Minimal allowed value, or `None` to impose no limits.
    :type min: int|float|None
    :param max: Maximal allowed value, or `None` to impose no limits.
    :type max: int|float|None
    """

    def __init__(self, min=None, max=None):
        self.min = min
        self.max = max

        # Name
        self.name = _(u'Clamp({min}..{max})').format(
            min=_(u'') if min is None else min,
            max=_(u'') if max is None else max
        )

    def __call__(self, v):
        # Clamp
        try:
            if self.min is not None and v < self.min:
                return self.min
            if self.max is not None and v > self.max:
                return self.max
        except TypeError:  # cannot compare
            raise Invalid(_(u'Value should be a number'), _(u'Number'), get_type_name(type(v)))

        # Ok
        return v


__all__ = ('Range', 'Clamp', )



""" Despite Good is modelled after Voluptuous and is highly compatible,
there still are differences that would definitely break your project.

If you're not ready for such a change -- `good.voluptuous` is the solution:
compatibility layer for switching from [voluptuous 0.8.5](https://github.com/alecthomas/voluptuous)
with 100% compatibility.

This is a drop-in replacement that passes all voluptuous unit-tests and hence should work perfectly.
Here's how to use it

```python
#from voluptuous import *  # no more
from good.voluptuous import *  # replacement

# .. and use it like before
```

Includes all the features and is absolutely compatible, except for the error message texts,
which became much more user-friendly :)

Migration steps:

1. Replace `voluptuous` imports with `good.voluptuous`
2. Run your application tests and see how it behaves
3. Module by module, replace `good.voluptuous` with just `good`, keeping the differences in mind.

Good luck! :)
"""

import good
import six
import os
from functools import wraps

def _convert_errors(func):
    """ Decorator to convert throws errors to Voluptuous format."""
    cast_Invalid = lambda e: Invalid(
        u"{message}, expected {expected}".format(
            message=e.message,
            expected=e.expected)
        if e.expected != u'-none-' else e.message,
        e.path,
        six.text_type(e))

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except good.SchemaError as e:
            raise SchemaError(six.text_type(e))
        except good.MultipleInvalid as ee:
            raise MultipleInvalid([cast_Invalid(e) for e in ee])
        except good.Invalid as e:
            # Since voluptuous throws MultipleInvalid almost always -- we follow the same pattern...
            raise MultipleInvalid([cast_Invalid(e)])
    return wrapper

from good.schema.util import Undefined  # it's not internal
UNDEFINED = Undefined()

# options for extra keys
PREVENT_EXTRA = 0
ALLOW_EXTRA = 1
REMOVE_EXTRA = 2

#region Errors

class Error(Exception): pass

class SchemaError(Error): pass

class Invalid(Error):
    def __init__(self, message, path=None, error_message=None):
        Error.__init__(self,  message)
        self.msg = message
        self.path = path or []
        self.error_message = error_message or message

    def __str__(self):
        path = ' @ data[%s]' % ']['.join(map(repr, self.path)) if self.path else ''
        return self.msg + path


class MultipleInvalid(Invalid):
    def __init__(self, errors=None):
        self.errors = errors[:] if errors else []
        e = errors[0]
        super(MultipleInvalid, self).__init__(e.msg, e.path, e.error_message)

    def __repr__(self):
        return 'MultipleInvalid(%r)' % self.errors

    def __str__(self):
        return str(self.errors[0])

#endregion


#region Markers

class Optional(good.Optional):
    def __init__(self, schema, msg=None):
        super(Optional, self).__init__(schema)

class Required(good.Required):
    def __init__(self, schema, msg=None, default=UNDEFINED):
        super(Required, self).__init__(schema)
        self.msg = msg
        # Voluptuous has `default` keyword here, but we use good.Default instead, so have to simulate
        self.default = default

    def execute(self, d, matches):
        # Simulate voluptuous `default=`
        if not matches and self.default is not UNDEFINED:
            return [(self.key_schema.schema, self.key_schema.schema, self.default)]

        try:
            return super(Required, self).execute(d, matches)
        except good.Invalid as e:
            # Replace the message (if required)
            if self.msg:
                e.message = self.msg
            raise e

from good import Extra
extra = Extra  # alias

#endregion

#region Schema

class Schema(object):
    #: Value conversion map for the `required` argument
    _required_arg_map = {
        True: Required,
        False: Optional
    }
    #: Value conversion map for the `extra` argument
    _extra_arg_map = {
        False: good.Reject,
        True: good.Allow,
        PREVENT_EXTRA: good.Reject,
        ALLOW_EXTRA: good.Allow,
        REMOVE_EXTRA: good.Remove,
    }

    @_convert_errors
    def __init__(self, schema, required=False, extra=PREVENT_EXTRA):
        self.schema = schema
        self.required = required
        self.extra = extra

        try:
            # It's a callable anyway, let it be here
            self._compiled = good.Schema(schema,
                                         default_keys=self._required_arg_map[required],
                                         extra_keys=self._extra_arg_map[extra])
        except good.SchemaError as e:
            raise SchemaError(e.message)

    @_convert_errors
    def __call__(self, data):
        return self._compiled(data)


#endregion

#region Helpers

class Object(good.Object):
    def __init__(self, schema, cls=UNDEFINED):
        # Convert argument
        if cls is UNDEFINED:
            cls = None
        # Proceed
        super(Object, self).__init__(schema, cls)

    def __call__(self, v):
        return super(Object, self).__call__(v)


def Msg(schema, msg):
    return good.Msg(schema, six.text_type(msg))

def message(default=None):
    message_decorator = good.message(six.text_type(default or u'invalid value'))
    # voluptuous produces a 2nd-level decorator which allows to override the message, good only allows Msg()
    def decorator(func):
        return lambda msg=None: _wrapMsg(message_decorator(func), msg)
    return decorator

def truth(f):
    return good.truth(u'not a valid value')(f)

#endregion

#region Validators

def _wrapMsg(v, msg=None):
    """ Wrap a validator in Msg, if `msg` was provided """
    return v if not msg else Msg(v, msg)

def Coerce(type, msg=None):
    return _wrapMsg(good.Coerce(type), msg)

def IsTrue():
    return good.Truthy()

def IsFalse():
    return good.Falsy()

def Boolean():
    return good.Boolean()

def Any(*validators, **kwargs):
    msg = kwargs.pop('msg', None)
    assert not kwargs, 'Sorry, Any() does not support Schema keyword arguments anymore'
    return _wrapMsg(good.Any(*validators), msg)

def All(*validators, **kwargs):
    msg = kwargs.pop('msg', None)
    assert not kwargs, 'Sorry, All() does not support Schema keyword arguments anymore'
    return _wrapMsg(good.All(*validators), msg)

def Match(pattern, msg=None):
    return good.Match(pattern, msg)

def Replace(pattern, substitution, msg=None):
    return good.Replace(pattern, substitution, msg)

def Url():
    return good.Url(['http', 'https', 'ftp'])

@message(u'not a file')
@truth
def IsFile(v):
    return os.path.isfile(v)

@message(u'not a directory')
@truth
def IsDir(v):
    return os.path.isdir(v)

@message(u'path does not exist')
@truth
def PathExists(v):
    return os.path.exists(v)

def Range(min=None, max=None, min_included=True, max_included=True, msg=None):
    # Alter min, max when the range is not inclusive
    if min is not None and not min_included:
        min += 1
    if max is not None and not max_included:
        max -= 1
    # Finish
    return _wrapMsg(good.Range(min, max), msg)

def Clamp(min=None, max=None, msg=None):
    return _wrapMsg(good.Clamp(min, max), msg)

def Length(min=None, max=None, msg=None):
    return _wrapMsg(good.Length(min, max), msg)

def In(container, msg=None):
    return _wrapMsg(good.In(container), msg)

# These are not callable in voluptuous
Lower = good.Lower()
Upper = good.Upper()
Capitalize = good.Capitalize()
Title = good.Title()

def DefaultTo(default_value, msg=None):
    # Completely useless!
    @wraps(DefaultTo)
    def f(v):
        if v is None:
            v = default_value
        return v
    return f

#endregion



from .. import Schema, Invalid, MultipleInvalid, Required, Optional
from ._base import ValidatorBase
from ..schema.util import get_literal_name


class Any(ValidatorBase):
    """ Try the provided schemas in order and use the first one that succeeds.

    This is the *OR* condition predicate: any of the schemas should match.
    [`Invalid`](#invalid) error is reported if neither of the schemas has matched.

    ```python
    from good import Schema, Any

    schema = Schema(Any(
        # allowed string constants
        'true', 'false',
        # otherwise coerce as a bool
        lambda v: 'true' if v else 'false'
    ))
    schema('true')  #-> 'true'
    schema(0)  #-> 'false'
    ```

    :param schemas: List of schemas to try.
    """

    def __init__(self, *schemas):
        # Compile
        self.compiled = tuple(Schema(schema) for schema in schemas)

        # Name
        self.name = _(u'Any({})').format(_(u'|'.join(x.name for x in self.compiled)))

    def __call__(self, v):
        # Try schemas in order
        for schema in self.compiled:
            try:
                return schema(v)
            except Invalid:
                pass

        # Nothing worked
        raise Invalid(_(u'Invalid value'))


class All(ValidatorBase):
    """ Value must pass all validators wrapped with `All()` predicate.

    This is the *AND* condition predicate: all of the schemas should match in order,
    which is in fact a composition of validators: `All(f,g)(value) = g(f(value))`.

    ```python
    from good import Schema, All, Range

    schema = Schema(All(
        # Must be an integer ..
        int,
        # .. and in the allowed range
        Range(0, 10)
    ))

    schema(1)  #-> 1
    schema(99)
    #-> Invalid: Not in range: expected 0..10, got 99
    ```

    :param schemas: List of schemas to apply.
    """

    def __init__(self, *schemas):
        # Compile
        self.compiled = tuple(Schema(schema) for schema in schemas)

        # Name
        self.name = _(u'All({})').format(_(u' & '.join(x.name for x in self.compiled)))

    def __call__(self, v):
        # Apply schemas in order and transform the value iteratively
        for schema in self.compiled:
            # Any failing schema will immediately throw an error
            v = schema(v)
        # Finished
        return v


class Neither(ValidatorBase):
    """ Value must not match any of the schemas.

    This is the *NOT* condition predicate: a value is considered valid if each schema has raised an error.

    ```python
    from good import Schema, All, Neither

    schema = Schema(All(
        # Integer
        int,
        # But not zero
        Neither(0)
    ))

    schema(1)  #-> 1
    schema(0)
    #-> Invalid: Value not allowed: expected Not(0), got 0
    ```

    :param schemas: List of schemas to check against.
    """

    def __init__(self, *schemas):
        # Compile
        self.compiled = tuple(Schema(schema) for schema in schemas)

        # Name
        self.name = _(u'Neither({})').format(_(u','.join(x.name for x in self.compiled)))

    def __call__(self, v):
        # Try schemas in order
        for schema in self.compiled:
            try:
                schema(v)
            except Invalid:
                pass  # error is okay
            else:
                raise Invalid(_(u'Value not allowed'), _(u'Not({})').format(schema.name), validator=schema.compiled.schema)

        # All ok
        return v


class Inclusive(ValidatorBase):
    """ `Inclusive` validates the defined inclusive group of mapping keys:
    if any of them was provided -- then all of them become required.

    This exists to support "sub-structures" within the mapping which only make sense if specified together.
    Since this validator works on the entire mapping, the best way is to use it together with the [`Entire`](#entire)
    marker:

    ```python
    from good import Schema, Entire, Inclusive

    schema = Schema({
        # Fields for all files
        'name': str,
        # Fields for images only
        Optional('width'): int,
        Optional('height'): int,
        # Now put a validator on the entire mapping
        Entire: Inclusive('width', 'height')
    })

    schema({'name': 'monica.jpg'})  #-> ok
    schema({'name': 'monica.jpg', 'width': 800, 'height': 600})  #-> ok
    schema({'name': 'monica.jpg', 'width': 800})
    #-> Invalid: Required key not provided: expected height, got -none-
    ```

    Note that `Inclusive` only supports literals.

    :param keys: List of mutually inclusive keys (literals).
    """

    def __init__(self, *keys):
        self.keys = set(keys)
        self.name = _(u'Inclusive({})').format(_(u',').join(keys))

    def __call__(self, d):
        # Check which keys are missing in the input mapping
        missing_keys = self.keys - set(d)

        # If none of the keys is specified in the input -- that's ok
        if missing_keys == self.keys:
            return d

        # If all keys were provided -- that's okay as well
        if not missing_keys:
            return d

        # Otherwise, we complain on every single key
        raise MultipleInvalid.if_multiple([
            Invalid(_(u'Required key not provided'), get_literal_name(key), _(u'-none-'), [key])
            for key in missing_keys
        ])


class Exclusive(ValidatorBase):
    """ `Exclusive` validates the defined exclusive group of mapping keys:
    if any of them was provided -- then none of the remaining keys can be used.

    This supports "sub-structures" with choice: if the user chooses a field from one of them --
    then he cannot use others.
    It works on the entire mapping and hence best to use with the [`Entire`](#entire) marker.

    By default, `Exclusive` requires the user to choose one of the options,
    but this can be overridden with [`Optional`](#optional) marker class given as an argument:

    ```python
    from good import Exclusive, Required, Optional

    # Requires either of them
    Exclusive('login', 'password')
    Exclusive(Required, 'login', 'password')  # the default

    # Requires either of them, or none
    Exclusive(Optional, 'login', 'password')
    ```

    Let's demonstrate with the API that supports multiple types of authentication,
    but requires the user to choose just one:

    ```python
    from good import Schema, Entire, Exclusive

    schema = Schema({
        # Authentication types: login+password | email+password
        Optional('login'): str,
        Optional('email'): str,
        'password': str,
        # Now put a validator on the entire mapping
        # that forces the user to choose
        Entire: Msg(  # also override the message
            Exclusive('login', 'email'),
            u'Choose one'
        )
    })

    schema({'login': 'kolypto', 'password': 'qwerty'})  #-> ok
    schema({'email': 'kolypto', 'password': 'qwerty'})  #-> ok
    schema({'login': 'a', 'email': 'b', 'password': 'c'})
    #-> MultipleInvalid:
    #->     Invalid: Choose one @ [login]: expected login|email, got login
    #->     Invalid: Choose one @ [email]: expected login|email, got email
    ```

    Note that `Exclusive` only supports literals.

    :param keys: List of mutually exclusive keys (literals).

        Can contain [`Required`](#required) or [`Optional`](#optional) marker classes,
        which defines the behavior when no keys are provided. Default is `Required`.
    """

    def __init__(self, *keys):
        keys = set(keys)

        # Required or Optional?
        try:
            # Optional
            keys.remove(Optional)
            self.require_mode = False
        except KeyError:
            # Required
            keys.discard(Required)
            self.require_mode = True

        # Name
        self.keys = set(keys)
        self.name = _(u'Exclusive({})').format(_(u',').join(sorted(self.keys)))

    def __call__(self, d):
        # Check which of the keys are provided
        provided_keys = self.keys & set(d)

        # None used
        if not provided_keys:
            if self.require_mode:
                # Required mode: fail
                raise Invalid(_(u'Choose one of the options'), provided=_(u'-none-'))
            else:
                # Optional mode: ok
                return d

        # One used: ok
        if len(provided_keys) == 1:
            return d

        # Multiple used
        raise Invalid(_(u'Choose one of the options, not multiple'), provided=_(u',').join(sorted(provided_keys)))



__all__ = ('Any', 'All', 'Neither', 'Inclusive', 'Exclusive')



import six


class ValidatorBase(object):
    """ Base for class-based validators """

    #: Validator name.
    #: Must be overridden in subclasses, and potentially hold the value
    name = u'???'

    def __call__(self, v):
        """ Do validation

        :param v: Input value
        :return: Sanitized value
        :raises Invalid: errors
        """
        raise NotImplementedError

    def __repr__(self):
        return self.name

    def __str__(self):
        return six.text_type(self).encode('utf8')

    def __unicode__(self):
        return self.name

    if six.PY3:
        __bytes__, __str__ = __str__, __unicode__



import six

from ._base import ValidatorBase
from .. import Invalid
from ..schema.util import get_callable_name, get_primitive_name, get_type_name


class Check(ValidatorBase):
    """ Use the provided boolean function as a validator and raise errors when it's `False`.

    ```python
    import os.path
    from good import Schema, Check

    schema = Schema(
        Check(os.path.isdir, u'Must be an existing directory'))
    schema('/')  #-> '/'
    schema('/404')
    #-> Invalid: Must be an existing directory: expected isDir(), got /404
    ```

    :param bvalidator: Boolean validator function
    :type bvalidator: callable
    :param message: Error message to report when `False`
    :type message: unicode
    :param expected: Expected value string representation, or `None` to get it from the wrapped callable
    :type expected: None|str|unicode
    """

    def __init__(self, bvalidator, message, expected):
        assert isinstance(message, six.text_type), 'Check() message must be a unicode string'
        assert isinstance(expected, six.text_type) or expected is None, 'Check() expected must be a unicode string'

        self.bvalidator = bvalidator
        self.name = get_callable_name(bvalidator)
        self.message = message
        self.expected = expected

    def __call__(self, v):
        #  Test with the boolean function
        if self.bvalidator(v):
            # Return value as is
            return v
        else:
            # If the boolean function reported False -- raise Invalid
            raise Invalid(self.message, self.expected)


class Truthy(ValidatorBase):
    """ Assert that the value is truthy, in the Python sense.

    This fails on all "falsy" values: `False`, `0`, empty collections, etc.

    ```python
    from good import Schema, Truthy

    schema = Schema(Truthy())

    schema(1)  #-> 1
    schema([1,2,3])  #-> [1,2,3]
    schema(None)
    #-> Invalid: Empty value: expected truthy(), got None
    ```
    """

    @classmethod
    def truthy(cls, v):
        return bool(v)

    def __init__(self):
        self.name = _(u'Truthy')

    def __call__(self, v):
        if not self.truthy(v):
            raise Invalid(u'Empty value', provided=get_primitive_name(v))
        return v


class Falsy(ValidatorBase):
    """ Assert that the value is falsy, in the Python sense.

    Supplementary to [`Truthy`](#truthy).
    """

    @classmethod
    def falsy(cls, v):
        return not bool(v)

    def __init__(self):
        self.name = _(u'Falsy')

    def __call__(self, v):
        if not self.falsy(v):
            raise Invalid(u'Non-empty value', provided=get_primitive_name(v))
        return v


class Boolean(ValidatorBase):
    """ Convert human-readable boolean values to a `bool`.

    The following values are supported:

    * `None`: `False`
    * `bool`: direct
    * `int`: `0` = `False`, everything else is `True`
    * `str`: Textual boolean values, compatible with [YAML 1.1 boolean literals](http://yaml.org/type/bool.html), namely:

            y|Y|yes|Yes|YES|n|N|no|No|NO|
            true|True|TRUE|false|False|FALSE|
            on|On|ON|off|Off|OFF

        [`Invalid`](#invalid) is thrown if an unknown string literal is provided.

    Example:

    ```python
    from good import Schema, Boolean

    schema = Schema(Boolean())

    schema(None)  #-> False
    schema(0)  #-> False
    schema(1)  #-> True
    schema(True)  #-> True
    schema(u'yes')  #-> True
    ```
    """
    #: Case-insensitive constants for boolean strings
    _true_values_ci  = (u'y', u'Y', u'yes', u'Yes', u'YES', u'true',  u'True',  u'TRUE',  u'on',  u'On',  u'ON' )
    _false_values_ci = (u'n', u'N', u'no',  u'No',  u'NO',  u'false', u'False', u'FALSE', u'off', u'Off', u'OFF')

    def __init__(self):
        self.name = _(u'Boolean')

    def __call__(self, v):
        # None
        if v is None:
            return False
        # Bool, Int
        elif isinstance(v, six.integer_types):
            return v != 0
        # Str
        elif isinstance(v, six.string_types):
            v = six.text_type(v)
            # Match
            if v in self._true_values_ci:
                return True
            elif v in self._false_values_ci:
                return False
            else:
                raise Invalid(_(u'Wrong boolean value'))
        # Other types
        else:
            raise Invalid(_(u'Wrong boolean value type'), provided=get_type_name(type(v)))



__all__ = ('Check', 'Truthy', 'Falsy', 'Boolean')



#! /usr/bin/env python

from __future__ import print_function, division

import voluptuous
import good

import itertools
from datetime import datetime
from random import choice, randrange


def generate_random_type(valid):
    """ Generate a random type and samples for it.

    :param valid: Generate valid samples?
    :type valid: bool
    :return: type, sample-generator
    :rtype: type, generator
    """
    type = choice(['int', 'str'])

    r = lambda: randrange(-1000000000, 1000000000)

    if type == 'int':
        return int,  (r() if valid else str(r()) for i in itertools.count())
    elif type == 'str':
        return str, (str(r()) if valid else r() for i in itertools.count())
    else:
        raise AssertionError('!')


def generate_random_schema(valid):
    """ Generate a random plain schema, and a sample generation function.

    :param valid: Generate valid samples?
    :type valid: bool
    :returns: schema, sample-generator
    :rtype: *, generator
    """
    schema_type = choice(['literal', 'type'])

    if schema_type == 'literal':
        type, gen = generate_random_type(valid)
        value = next(gen)
        return value, (value if valid else None for i in itertools.count())
    elif schema_type == 'type':
        return generate_random_type(valid)
    else:
        raise AssertionError('!')


def generate_dict_schema(size, valid):
    """ Generate a schema dict of size `size` using library `lib`.

    In addition, it returns samples generator

    :param size: Schema size
    :type size: int
    :param samples: The number of samples to generate
    :type samples: int
    :param valid: Generate valid samples?
    :type valid: bool
    :returns
    """

    schema = {}
    generator_items = []

    # Generate schema
    for i in range(0, size):
        while True:
            key_schema,   key_generator   = generate_random_schema(valid)
            if key_schema not in schema:
                break
        value_schema, value_generator = generate_random_schema(valid)

        schema[key_schema] = value_schema
        generator_items.append((key_generator, value_generator))

    # Samples
    generator = ({next(k_gen): next(v_gen) for k_gen, v_gen in generator_items} for i in itertools.count())

    # Finish
    return schema, generator



if __name__ == '__main__':
    import sys
    import argparse
    from collections import defaultdict

    parser = argparse.ArgumentParser(prog='Performance')
    parser.add_argument('samples', type=int, help='The number of samples to test with')
    parser.add_argument('size_min', type=int, help='Min dictionary size')
    parser.add_argument('size_max', type=int, help='Max dictionary size')
    args = parser.parse_args()

    # Test on both valid and invalid schemas
    results = defaultdict(list)
    best_for_size = defaultdict(float)
    for valid in (True, False):

        # Generate schemas of different size
        dictionaries = []
        for size in range(args.size_min, args.size_max + 1):
            # Generate samples
            schema, gen = generate_dict_schema(size, valid)
            samples = list(sample for i, sample in zip(range(0, args.samples), gen))

            dictionaries.append((
                size,
                schema,
                samples
            ))

        # Iterate over libraries
        for lib in (good, voluptuous):
            lib_name = lib.__name__

            # Generate schemas of different size
            for size, schema, samples in dictionaries:
                compiled_schema = lib.Schema(schema.copy())

                # Now do validation
                start = datetime.utcnow()
                for sample in samples:
                    try:
                        compiled_schema(sample)
                    except lib.Invalid as e:
                        # Ignore errors
                        pass
                stop = datetime.utcnow()

                # Results
                spent_time = (stop - start).total_seconds()
                validations_per_second = len(samples) / spent_time

                # Save
                results[valid, lib_name].append(dict(
                    size=size,
                    sec=spent_time,
                    vps=validations_per_second
                ))

                # Best
                best_for_size[valid, size] = max(
                    best_for_size[valid, size],
                    validations_per_second)

    # Print dataset
    for (valid, lib_name), stats_list in sorted(results.items()):
        # Dataset header
        print('"{lib} ({valid})"'.format(
            lib=lib_name,
            valid='Valid' if valid else 'Invalid',
        ))

        # Values
        print("#size  time  vps")
        for stat in stats_list:
            print('{size: 5d} {sec: 4.2f} {vps: 10.2f}'.format(**stat))
        print('\n')  # split datasets

        # Calculate averages
    for (valid, lib_name), stats_list in sorted(results.items()):
        vps = sum(x['vps'] for x in stats_list) / len(stats_list)
        print('AVG:{lib:<12} {valid:<8} {vps: 10.2f}'.format(
            lib=lib_name,
            valid='Valid' if valid else 'Invalid',
            vps=vps
        ), file=sys.stderr)



import six
from functools import wraps
import re

from ._base import ValidatorBase
from .. import Invalid
from ..schema.util import get_type_name


def stringmethod(func):
    """ Validator factory which call a single method on the string. """
    method_name = func()

    @wraps(func)
    def factory():
        def validator(v):
            if not isinstance(v, six.string_types):
                raise Invalid(_(u'Not a string'), get_type_name(six.text_type), get_type_name(type(v)))
            return getattr(v, method_name)()
        return validator
    return factory


@stringmethod
def Lower():
    """ Casts the provided string to lowercase, fails is the input value is not a string.

    Supports both binary and unicode strings.

    ```python
    from good import Schema, Lower

    schema = Schema(Lower())

    schema(u'ABC')  #-> u'abc'
    schema(123)
    #-> Invalid: Not a string: expected String, provided Integer number
    ```
    """
    return 'lower'


@stringmethod
def Upper():
    """ Casts the input string to UPPERCASE. """
    return 'upper'


@stringmethod
def Capitalize():
    """ Capitalizes the input string. """
    return 'capitalize'


@stringmethod
def Title():
    """ Casts The Input String To Title Case """
    return 'title'


class Match(ValidatorBase):
    """ Validate the input string against a regular expression.

    ```python
    from good import Schema, Match

    schema = Schema(All(
        unicode,
        Match(r'^0x[A-F0-9]+$', 'hex number')
    ))

    schema('0xDEADBEEF')  #-> '0xDEADBEEF'
    schema('0x')
    #-> Invalid: Wrong format: expected hex number, got 0xDEADBEEF
    ```

    :param pattern: RegExp pattern to match with: a string, or a compiled pattern
    :type pattern: str|_SRE_Pattern
    :param message: Error message override
    :type message: unicode
    :param expected: Textual representation of what's expected from the user
    :type expected: unicode
    """

    def __init__(self, pattern, message=None, expected=None):
        self.rex = re.compile(pattern)  # accepts compiled patterns as well
        self.name = expected or _(u'(special format)')
        self.message = message or _(u'Wrong format')

    def __call__(self, v):
        try:
            # Try to match
            match = self.rex.match(v)
        except TypeError:
            # Wrong type
            raise Invalid(_(u'Wrong value type'), u'String', get_type_name(type(v)))

        # Matched?
        if not match:
            raise Invalid(self.message)
        else:
            return v


class Replace(Match):
    """ RegExp substitution.

    ```python
    from good import Schema, Replace

    schema = Schema(Replace(
        # Grab domain name
        r'^https?://([^/]+)/.*'
        # Replace
        r'\1',
        # Tell the user that we're expecting a URL
        u'URL'
    ))

    schema('http://example.com/a/b/c')  #-> 'example.com'
    schema('user@example.com')
    #-> Invalid: Wrong format: expected URL, got user@example.com
    ```

    :param pattern: RegExp pattern to match with: a string, or a compiled pattern
    :type pattern: str|_SRE_Pattern
    :param repl: Replacement pattern.

        Backreferences are supported, just like in the [`re`](https://docs.python.org/2/library/re.html) module.

    :type repl: unicode
    :param message: Error message override
    :type message: unicode|None
    :param expected: Textual representation of what's expected from the user
    :type expected: unicode
    """

    def __init__(self, pattern, repl, message=None, expected=None):
        super(Replace, self).__init__(pattern, message, expected)
        self.repl = repl

    def __call__(self, v):
        try:
            # Try to match
            v, n_subs = self.rex.subn(self.repl, v)
        except TypeError:
            # Wrong type
            raise Invalid(_(u'Wrong value type'), u'String', get_type_name(type(v)))

        # Matched?
        if not n_subs:
            raise Invalid(self.message)
        else:
            return v


class Url(ValidatorBase):
    """ Validate a URL, make sure it's in the absolute format, including the protocol.

    ```python
    from good import Schema, Url

    schema = Schema(Url('https'))

    schema('example.com')  #-> 'https://example.com'
    schema('http://example.com')  #-> 'http://example.com'
    ```

    :param protocols: List of allowed protocols.

        If no protocol is provided by the user -- the first protocol is used by default.

    :type protocols: str|list[str]
    """

    _url_rex = r'^' \
               r'(?:' r'(?P<scheme>[^:]+)' r'://?)?' \
               r'(?:' r'(?P<auth>[^@/]+(?::[^@/]*)?)' r'@)?' \
               r'(?P<host>[^/@:]+)' \
               r'(?:' r':(?P<port>\d+)' r')?' \
               r'(?:' r'(?P<path>/.*)' r')?' \
               r'$'

    name = u'URL'

    def __init__(self, protocols=('http', 'https')):
        self.protocols = tuple(x.lower()
                               for x in ((protocols,)
                                         if isinstance(protocols, six.string_types) else
                                         tuple(protocols)))

        self.rex = re.compile(self._url_rex)

    def __call__(self, v):
        # Match
        try:
            match = self.rex.match(v)
        except TypeError:
            raise Invalid(_(u'Wrong URL value type'), u'String', get_type_name(type(v)))

        # Matched?
        if not match:
            raise Invalid(_(u'Wrong URL format'))

        try:
            # Prepare
            parts = match.groupdict()
            if not parts['scheme']:
                parts['scheme'] = self.protocols[0]

            # Validate
            if parts['scheme'].lower() not in self.protocols:
                raise Invalid(u'Protocol not allowed', _(u',').join(self.protocols), six.text_type(parts['scheme']))
            if '.' not in parts['host']:
                raise Invalid(u'Incorrect domain name')

            # Combine back again
            return six.moves.urllib.parse.urlunsplit((
                parts['scheme'],
                  ('{auth}@'.format(**parts) if parts['auth'] else '')
                + parts['host']
                + (':{port}'.format(**parts) if parts['port'] else ''),
                parts['path'] or '/',
                None,
                None
            ))
        except Exception as e:
            if isinstance(e, Invalid):
                raise
            else:
                # Other error types are not expected, so reraise them
                raise RuntimeError('{}: {}'.format(type(e).__name__, e))


class Email(Match):
    """ Validate that a value is an e-mail address.

    This simply tests for the presence of the '@' sign, surrounded by some characters.

    ```python
    from good import Email

    schema = Schema(Email())

    schema('user@example.com')  #-> 'user@example.com'
    schema('user@localhost')  #-> 'user@localhost'
    schema('user')
    #-> Invalid: Wrong e-mail: expected E-Mail, got user
    ```
    """

    _rex = re.compile(r'.+@.+')

    def __init__(self):
        super(Email, self).__init__(self._rex, u'Wrong E-Mail', u'E-Mail')


__all__ = ('Lower', 'Upper', 'Capitalize', 'Title', 'Match', 'Replace', 'Url', 'Email')



import six

from . import markers, signals
from .errors import SchemaError, Invalid, MultipleInvalid
from .util import get_type_name, get_literal_name, get_callable_name,  const, primitive_type


def Identity(v):
    """ Identity function.

    Special `identity` function which matches everything.
    Is primarily used for the `Extra` marker, as well as all other markers specified as classes.
    """
    return v
Identity.name = _(u'*')  # Set a name on it (for repr())


class CompiledSchema(object):
    """ Schema compiler.

    Converts a Schema into a callable, recursively.

    :param schema: Schema to use for validation
    :param path: Path to this schema
    :param default_keys: Default dictionary keys behavior (marker class)
    :param extra_keys: Default extra keys behavior (schema | marker class)
    :param matcher: Compile a "Matcher" schema: instead of throwing exceptions, it returns a boolean which indicates whether the value matches.

            This is used with mapping validation: a "matcher" is a lightweight alternative to CompiledSchema which economizes exceptions in favor of just returning booleans.

            Note that some values cannot be matchers: e.g. callables, which can typecast dictionary keys.
    """

    def __init__(self, schema, path, default_keys, extra_keys, matcher=False):
        assert issubclass(default_keys, markers.Marker), '`default_keys` value must be a Marker'

        self.path = path
        self.schema = schema
        self.default_keys = default_keys
        self.extra_keys = extra_keys
        self.matcher = matcher

        # Compile
        self.name = None
        self.compiled_type = None
        self.compiled = self.compile_schema(self.schema)

        assert self.compiled_type is not None, 'Compiler did not set a schema `compiled_type`'
        assert isinstance(self.name, six.text_type), 'Compiler did not set a valid schema name: {!r} (must be unicode)'.format(self.name)

    def __call__(self, value):
        """ Validate value against the compiled schema

        :param value: The value to validate
        :return: Sanitized value

            However, if the schema was compiled as "Matcher" (matcher=True) -- it returns:

            ( is-okay, sanitized-value )

            <is-okay> tells whether the matcher has matched, and <sanitized-value> is the final value (if it matched)
        """
        return self.compiled(value)

    def __repr__(self):
        return '{cls}{matcher}({0.schema!r}, ' \
               '{0.path!r}, ' \
               '{default_keys}, ' \
               '{extra_keys})' \
            .format(
                self,
                cls=type(self).__name__,
                matcher=':Matcher' if self.matcher else '',
                default_keys=self.default_keys.__name__,
                extra_keys=self.extra_keys.__name__ if isinstance(self.extra_keys, type) else repr(self.extra_keys)
            )

    def __unicode__(self):
        return self.name

    if six.PY3:
        __str__ = __unicode__

    @property
    def supports_undefined(self):
        """ Test whether this schema supports Undefined.

        A Schema that supports `Undefined`, when given `Undefined`, should return some value (other than `Undefined`)
        without raising errors.

        This is designed to support a very special case like that:

        ```python
        Schema(Default(0)).supports_undefined  #-> True
        ```

        This way a validator can declare that it has a default in case no value was provided,
        and this case happens when:

        1. A [`Required`](#required) mapping key was not provided, and it's mapped to `Default()`
        2. .. no more supported cases. Yet.

        :rtype: bool
        """
        # Test
        try:
            yes = self(const.UNDEFINED) is not const.UNDEFINED
        except Invalid:
            yes = False

        # Remember (lame @cached_property)
        self.__dict__['supports_undefined'] = yes
        return yes

    #region Compilation Utils

    @classmethod
    def get_schema_type(cls, schema):
        """ Get schema type for the argument

        :param schema: Schema to analyze
        :return: COMPILED_TYPE constant
        :rtype: str|None
        """
        schema_type = type(schema)

        # Marker
        if issubclass(schema_type, markers.Marker):
            return const.COMPILED_TYPE.MARKER
        # Marker Type
        elif issubclass(schema_type, six.class_types) and issubclass(schema, markers.Marker):
            return const.COMPILED_TYPE.MARKER
        # CompiledSchema
        elif isinstance(schema, CompiledSchema):
            return const.COMPILED_TYPE.SCHEMA
        else:
            return primitive_type(schema)

    @property
    def priority(self):
        """ Get priority for this Schema.

        Used to sort mapping keys

        :rtype: int
        """
        # Markers have priority set on the class
        if self.compiled_type == const.COMPILED_TYPE.MARKER:
            return self.compiled.priority

        # Other types have static priority
        return const.compiled_type_priorities[self.compiled_type]

    @classmethod
    def sort_schemas(cls, schemas_list):
        """ Sort the provided list of schemas according to their priority.

        This also supports markers, and markers of a single type are also sorted according to the priority of the wrapped schema.

        :type schemas_list: list[CompiledSchema]
        :rtype: list[CompiledSchema]
        """
        return sorted(schemas_list,
                      key=lambda x: (
                          # Top-level priority:
                          # priority of the schema itself
                          x.priority,
                          # Second-level priority (for markers of the common type)
                          # This ensures that Optional(1) always goes before Optional(int)
                          x.compiled.key_schema.priority if x.compiled_type == const.COMPILED_TYPE.MARKER else 0
                      ), reverse=True)

    def sub_compile(self, schema, path=None, matcher=False):
        """ Compile a sub-schema

        :param schema: Validation schema
        :type schema: *
        :param path: Path to this schema, if any
        :type path: list|None
        :param matcher: Compile a matcher?
        :type matcher: bool
        :rtype: CompiledSchema
        """
        return type(self)(
            schema,
            self.path + (path or []),
            self.default_keys,
            self.extra_keys,
            matcher
        )

    def Invalid(self, message, expected):
        """ Helper for Invalid errors.

        Typical use:

        err_type = self.Invalid(_(u'Message'), self.name)
        raise err_type(<provided-value>)

        Note: `provided` and `expected` are unicode-typecasted automatically

        :type message: unicode
        :type expected: unicode
        """
        def InvalidPartial(provided, path=None, **info):
            """ Create an Invalid exception

            :type provided: unicode
            :type path: list|None
            :rtype: Invalid
            """
            return Invalid(
                message,
                expected, #six.text_type(expected),  # -- must be unicode
                provided, #six.text_type(provided),  # -- must be unicode
                self.path + (path or []),
                self.schema,
                **info
            )
        return InvalidPartial

    #endregion

    #region Compilation Procedure

    def get_schema_compiler(self, schema):
        """ Get compiler method for the provided schema

        :param schema: Schema to analyze
        :return: Callable compiled
        :rtype: callable|None
        """
        # Schema type
        schema_type = self.get_schema_type(schema)
        if schema_type is None:
            return None

        # Compiler
        compilers = {
            const.COMPILED_TYPE.LITERAL: self._compile_literal,
            const.COMPILED_TYPE.TYPE: self._compile_type,
            const.COMPILED_TYPE.SCHEMA: self._compile_schema,
            const.COMPILED_TYPE.CALLABLE: self._compile_callable,
            const.COMPILED_TYPE.ITERABLE: self._compile_iterable,
            const.COMPILED_TYPE.MAPPING: self._compile_mapping,
            const.COMPILED_TYPE.MARKER: self._compile_marker,
        }

        return compilers[schema_type]

    def compile_schema(self, schema):
        """ Compile the current schema into a callable validator

        :return: Callable validator
        :rtype: callable
        :raises SchemaError: Schema compilation error
        """
        compiler = self.get_schema_compiler(schema)

        if compiler is None:
            raise SchemaError(_(u'Unsupported schema data type {!r}').format(type(schema).__name__))

        return compiler(schema)

    def _compile_literal(self, schema):
        """ Compile literal schema: type and value matching """
        # Prepare self
        self.compiled_type = const.COMPILED_TYPE.LITERAL
        self.name = get_literal_name(schema)

        # Error partials
        schema_type = type(schema)
        err_type  = self.Invalid(_(u'Wrong value type'), get_type_name(schema_type))
        err_value = self.Invalid(_(u'Invalid value'), self.name)

        # Matcher
        if self.matcher:
            def match_literal(v):
                return type(v) == schema_type and v == schema, v
            return match_literal

        # Validator
        def validate_literal(v):
            # Type check
            if type(v) != schema_type:
                # expected=<type>, provided=<type>
                raise err_type(get_type_name(type(v)))
            # Equality check
            if v != schema:
                # expected=<value>, provided=<value>
                raise err_value(get_literal_name(v))
            # Fine
            return v
        return validate_literal

    def _compile_type(self, schema):
        """ Compile type schema: plain type matching """
        # Prepare self
        self.compiled_type = const.COMPILED_TYPE.TYPE
        self.name = get_type_name(schema)

        # Error partials
        err_type = self.Invalid(_(u'Wrong type'), self.name)

        # Type check function
        if six.PY2 and schema is basestring:
            # Relaxed rule for Python2 basestring
            typecheck = lambda v: isinstance(v, schema)
        else:
            # Strict type check for everything else
            typecheck = lambda v: type(v) == schema

        # Matcher
        if self.matcher:
            def match_type(v):
                return typecheck(v), v
            return match_type

        # Validator
        def validate_type(v):
            # Type check
            if not typecheck(v):
                # expected=<type>, provided=<type>
                raise err_type(get_type_name(type(v)))
            # Fine
            return v

        return validate_type

    def _compile_schema(self, schema):
        """ Compile another schema """
        assert self.matcher == schema.matcher

        self.name = schema.name
        self.compiled_type = schema.compiled_type

        return schema.compiled

    def _compile_callable(self, schema):
        """ Compile callable: wrap exceptions with correct paths """
        # Prepare self
        self.compiled_type = const.COMPILED_TYPE.CALLABLE
        self.name = get_callable_name(schema)

        # Error utils
        enrich_exception = lambda e, value: e.enrich(
            expected=self.name,
            provided=get_literal_name(value),
            path=self.path,
            validator=schema)

        # Validator
        def validate_with_callable(v):
            try:
                # Try this callable
                return schema(v)
            except Invalid as e:
                # Enrich & re-raise
                enrich_exception(e, v)
                raise
            except const.transformed_exceptions as e:
                message = _(u'{message}').format(
                    Exception=type(e).__name__,
                    message=six.text_type(e))
                e = Invalid(message)
                raise enrich_exception(e, v)

        # Matcher
        if self.matcher:
            def match_with_callable(v):
                try:
                    return True, validate_with_callable(v)
                except Invalid:
                    return False, v
            return match_with_callable

        return validate_with_callable

    def _compile_iterable(self, schema):
        """ Compile iterable: iterable of schemas treated as allowed values """
        # Compile each member as a schema
        schema_type = type(schema)
        schema_subs = tuple(map(self.sub_compile, schema))

        # Prepare self
        self.compiled_type = const.COMPILED_TYPE.ITERABLE
        self.name = _(u'{iterable_cls}[{iterable_options}]').format(
            iterable_cls=get_type_name(schema_type),
            iterable_options=_(u'|').join(x.name for x in schema_subs)
        )

        # Error partials
        err_type = self.Invalid(_(u'Wrong value type'), get_type_name(schema_type))
        err_value = self.Invalid(_(u'Invalid value'), self.name)

        # Validator
        def validate_iterable(l):
            # Type check
            if not isinstance(l, schema_type):
                # expected=<type>, provided=<type>
                raise err_type(provided=get_type_name(type(l)))

            # Each `v` member should match to any `schema` member
            errors = []  # Errors for every value
            values = []  # Sanitized values
            for value_index, value in list(enumerate(l)):
                # Walk through schema members and test if any of them match
                for value_schema in schema_subs:
                    try:
                        # Try to validate
                        values.append(value_schema(value))
                        break  # Success!
                    except signals.RemoveValue:
                        # `value_schema` commanded to drop this value
                        break
                    except Invalid:
                        # Ignore errors and hope other members will succeed better
                        pass
                else:
                    errors.append(err_value(get_literal_name(value), path=[value_index]))

            # Errors?
            if errors:
                raise MultipleInvalid.if_multiple(errors)

            # Typecast and finish
            return schema_type(values)

        # Matcher
        if self.matcher:
            return self._compile_callable(validate_iterable)  # Stupidly use it as callable

        return validate_iterable

    def _compile_marker(self, schema):
        """ Compile marker: sub-schema with special type """
        # Prepare self
        self.compiled_type = const.COMPILED_TYPE.MARKER

        # If this marker is not instantiated -- do it with an identity callable which is valid for everything
        if issubclass(type(schema), six.class_types):
            schema = schema(Identity)  \
                .on_compiled(name=Identity.name)  # Set a special name on it

        # Compile Marker's schema
        key_schema = self.sub_compile(schema.key, matcher=self.matcher)
        schema.on_compiled(key_schema=key_schema, name=key_schema.name)

        # Marker is a callable
        self.name = schema.name
        return schema

    def _compile_mapping(self, schema):
        """ Compile mapping: key-value matching """
        assert not self.matcher, 'Mappings cannot be matchers'

        # This stuff is tricky, but thankfully, I like comments :)

        # Set default Marker on all keys.
        # This makes sure that all keys will still become markers, and hence have the correct default behavior
        schema = {self.default_keys(k) if self.get_schema_type(k) != const.COMPILED_TYPE.MARKER else k: v
                  for k, v in schema.items()}

        # Add `Extra`
        # Every Schema implicitly has an `Extra` that defaults to `extra_keys`.
        # Note that this is the only place in the code where Marker behavior is hardcoded :)
        schema.setdefault(markers.Extra, self.extra_keys)

        # Compile both keys & values as schemas.
        # Key schemas are compiled as "Matchers" for performance.
        compiled = {self.sub_compile(key, matcher=True): self.sub_compile(value)
                    for key, value in schema.items()}

        # Notify Markers that they were compiled.
        # _compile_marker() has already done part of the job: it only specified `key_schema`.
        # Here we let the Marker know its `value_schema` as well.
        for key_schema, value_schema in compiled.items():
            if key_schema.compiled_type == const.COMPILED_TYPE.MARKER:
                key_schema.compiled.on_compiled(value_schema=value_schema, as_mapping_key=True)

        # Sort key schemas for matching.

        # Since various schema types have different priority, we need to sort these accordingly.
        # Then, literals match before types, and Markers can define execution order using `priority`.
        # For instance, Remove() should be called first (before any validation takes place),
        # while Extra() should be checked last so it catches all extra keys that did not match other key schemas.

        # In addition, since mapping keys are mostly literals, we want direct matching instead of the costly function calls.
        # Hence, remember which of them are literals or 'catch-all' markers.
        is_literal  = lambda key_schema: key_schema.compiled.key_schema.compiled_type == const.COMPILED_TYPE.LITERAL
        is_identity = lambda identity: key_schema.compiled.key_schema is Identity

        compiled = [ (key_schema, compiled[key_schema], is_literal(key_schema), is_identity(key_schema))
                     for key_schema in self.sort_schemas(compiled.keys())]
        ''' :var  compiled: Sorted list of CompiledSchemas: (key-schema, value-schema, is-literal),
            :type compiled: list[CompiledSchema, CompiledSchema, bool]
        '''

        # Prepare self
        self.compiled_type = const.COMPILED_TYPE.MAPPING
        self.name = _(u'{mapping_cls}[{mapping_keys}]').format(
            mapping_cls=get_type_name(type(schema)),
            mapping_keys=_(u',').join(key_schema.name for key_schema, value_schema, is_literal, is_identity in compiled)
        )

        # Error partials
        schema_type = type(schema)
        err_type = self.Invalid(_(u'Wrong value type'), get_type_name(schema_type))

        # Validator
        def validate_mapping(d):
            # Type check
            if not isinstance(d, schema_type):
                # expected=<type>, provided=<type>
                raise err_type(provided=get_type_name(type(d)))

            # For each schema key, pick matching input key-value pairs.
            # Since we always have Extra which is a catch-all -- this will always result into a full input coverage.
            # Also, key schemas are sorted according to the priority, we're handling each set of matching keys in order.

            errors = []  # Collect errors on the fly
            d_keys = set(d.keys())  # Make a copy of dict keys for destructive iteration

            for key_schema, value_schema, is_literal, is_identity in compiled:
                # First, collect matching (key, value) pairs for the `key_schema`.
                # Note that `key_schema` can change the value (e.g. `Coerce(int)`), so for every key
                # we store both the initial value (`input-key`) and the sanitized value (`sanitized-key`).
                # This results into a list of triples: [(input-key, sanitized-key, input-value), ...].

                matches = []

                if is_literal:  # (short-circuit for literals)
                    # Since mapping keys are mostly literals --
                    # save some iterations & function calls in favor of direct matching,
                    # which introduces a HUGE performance improvement
                    k = key_schema.schema.key  # get the literal from the marker
                    if k in d:
                        # (See comments below)
                        matches.append(( k, k, d[k] ))
                        d_keys.remove(k)
                elif is_identity:  # (short-circuit for Marker(Identity))
                    # When this value is an identity function -- we plainly add all keys to it.
                    # This is to short-circuit catch-all markers like `Extra`, which, being executed last,
                    # just gets all remaining keys.
                    matches.extend((k, k, d[k]) for k in d_keys)
                    d_keys = None  # empty it since we've processed everything
                elif d_keys:
                    # For non-literal schemas we have to walk all input keys
                    # and detect those that match the current `key_schema`.
                    # In contrast to literals, such keys may have multiple matches (e.g. `{ int: 1 }`).

                    # Note that this condition branch includes the logic from the short-circuited logic implemented above,
                    # but is less performant.

                    for k in tuple(d_keys):
                        # Exec key schema on the input key.
                        # Since all key schemas are compiled as matchers -- we get a tuple (key-matched, sanitized-key)

                        okay, sanitized_k = key_schema(k)

                        # If this key has matched -- append it to the list of matches for the current `key_schema`.
                        # Also, remove the key from the original input so it does not match any other key schemas
                        # with lower priorities.
                        if okay:
                            matches.append(( k, sanitized_k, d[k] ))
                            d_keys.remove(k)

                # Now, having a `key_schema` and a list of matches for it, do validation.
                # If the key is a marker -- execute the marker first so it has a chance to modify the input,
                # and then proceed with value validation.

                # Execute Marker first.
                if key_schema.compiled_type == const.COMPILED_TYPE.MARKER:
                    # Note that Markers can raise errors as well.
                    # Since they're compiled - all marker errors are raised as `Invalid`.
                    try:
                        matches = key_schema.compiled.execute(d, matches)
                    except Invalid as e:
                        # Add marker errors to the list of Invalid reports for this schema.
                        # Using enrich(), we're also setting `path` prefix, and other info known at this step.
                        errors.append(e.enrich(
                            # Markers are responsible to set `expected`, `provided`, `validator`
                            expected=key_schema.name,
                            provided=None,  # Marker's required to set that
                            path=self.path,
                            validator=key_schema.compiled
                        ))
                        # If a marker raised an error -- the (key, value) pair is already Invalid, and no
                        # further validation is required.
                        continue

                # Proceed with validation.
                # Now, we validate values for every (key, value) pairs in the current list of matches,
                # and rebuild the mapping.
                for k, sanitized_k, v in matches:
                    try:
                        # Execute the value schema and store it into the rebuilt mapping
                        # using the sanitized key, which might be different from the original key.
                        d[sanitized_k] = value_schema(v)

                        # Remove the original key in case `key_schema` has transformed it.
                        if k != sanitized_k:
                            del d[k]
                    except signals.RemoveValue:
                        # `value_schema` commanded to drop this value
                        del d[k]
                    except Invalid as e:
                        # Any value validation errors are appended to the list of Invalid reports for the schema
                        # enrich() adds more info on the collected errors.
                        errors.append(e.enrich(
                            expected=value_schema.name,
                            provided=get_literal_name(v),
                            path=self.path + [k],
                            validator=value_schema
                        ))

            assert not d_keys, 'Keys must be empty after destructive iteration. Remainder: {!r}'.format(d_keys)

            # Errors?
            if errors:
                # Note that we did not care about whether a sub-schema raised a single Invalid or MultipleInvalid,
                # since MultipleInvalid will flatten the list for us.
                raise MultipleInvalid.if_multiple(errors)

            # Finish
            return d

        return validate_mapping

    #endregion



""" Misc utilities """

import six
import collections


class Undefined(object):
    """ Special singleton object to represent the case when no value was provided.

    This value is never equal to anything and always returns False for any attempts to typecheck it:
    this makes sure it will never match any condition.
    """

    _instance = None

    def __new__(cls):
        # Singleton
        if cls._instance is None:
            cls._instance = super(Undefined, cls).__new__(cls)
        return cls._instance

    def __coerce__(self, other):
        return None

    def __eq__(self, other):
        return False

    def __repr__(self):
        return '<Undefined>'


__type_names = direct = {
    None:             _(u'None'),
    type(None):       _(u'None'),
    bool:       _(u'Boolean'),
    float:      _(u'Fractional number'),
    complex:    _(u'Complex number'),
    six.text_type:    _(u'String'),
    six.binary_type:  _(u'Binary String'),
    tuple:      _(u'Tuple'),
    list:       _(u'List'),
    set:        _(u'Set'),
    frozenset:  _(u'Frozen Set'),
    dict:       _(u'Dictionary'),
}


def register_type_name(t, name):
    """ Register a human-friendly name for the given type. This will be used in Invalid errors

    :param t: The type to register
    :type t: type
    :param name: Name for the type
    :type name: unicode
    """
    assert isinstance(t, type)
    assert isinstance(name, unicode)
    __type_names[t] = name


def get_literal_name(v):
    """ Get a human-friendly name for the given literal.

    :param v: Value
    :type v: *
    :rtype: unicode
    """
    return six.text_type(v)


def get_type_name(t):
    """ Get a human-friendly name for the given type.

    :type t: type|None
    :rtype: unicode
    """
    # Lookup in the mapping
    try:
        return __type_names[t]
    except KeyError:
        # Specific types
        if issubclass(t, six.integer_types):
            return _(u'Integer number')

        # Get name from the Type itself
        return six.text_type(t.__name__).capitalize()


def get_callable_name(c):
    """ Get a human-friendly name for the given callable.

    :param c: The callable to get the name for
    :type c: callable
    :rtype: unicode
    """
    if hasattr(c, 'name'):
        return six.text_type(c.name)
    elif hasattr(c, '__name__'):
        return six.text_type(c.__name__) + u'()'
    else:
        return six.text_type(c)


def get_primitive_name(schema):
    """ Get a human-friendly name for the given primitive.

    :param schema: Schema
    :type schema: *
    :rtype: unicode
    """
    try:
        return {
            const.COMPILED_TYPE.LITERAL: six.text_type,
            const.COMPILED_TYPE.TYPE: get_type_name,
            const.COMPILED_TYPE.CALLABLE: get_callable_name,
            const.COMPILED_TYPE.ITERABLE: lambda x: _(u'{type}[{content}]').format(type=get_type_name(list), content=_(u'...') if x else _(u'-')),
            const.COMPILED_TYPE.MAPPING:  lambda x: _(u'{type}[{content}]').format(type=get_type_name(dict), content=_(u'...') if x else _(u'-')),
        }[primitive_type(schema)](schema)
    except KeyError:
        return six.text_type(repr(schema))


class const:
    """ Misc constants """

    #: Undefined singleton
    UNDEFINED = Undefined()

    #: Types that are treated as literals
    literal_types = six.integer_types + (six.text_type, six.binary_type) + (bool, float, complex, object, type(None))

    #: Exception classes that are transformed into Invalid when thrown by a callable
    transformed_exceptions = (AssertionError, TypeError, ValueError,)


    class COMPILED_TYPE:
        """ Compiled schema types """

        LITERAL = 'literal'
        TYPE = 'type'
        SCHEMA = 'schema'
        CALLABLE = 'callable'

        ITERABLE = 'iterable'
        MAPPING = 'mapping'
        MARKER = 'marker'

    #: Priorities for compiled types
    #: This is used for mappings to determine the sequence with which the keys are processed
    #: See _schema_priority()

    compiled_type_priorities = {
        COMPILED_TYPE.LITERAL:   100,
        COMPILED_TYPE.TYPE:      50,
        COMPILED_TYPE.SCHEMA:     0,
        COMPILED_TYPE.CALLABLE:   0,
        COMPILED_TYPE.ITERABLE:   0,
        COMPILED_TYPE.MAPPING:    0,
        COMPILED_TYPE.MARKER:   None,  # Markers have their own priorities
    }


def primitive_type(schema):
    """ Get schema type for the primitive argument.

    Note: it does treats markers & schemas as callables!

    :param schema: Value of a primitive type
    :type schema: *
    :return: const.COMPILED_TYPE.*
    :rtype: str|None
    """
    schema_type = type(schema)

    # Literal
    if schema_type in const.literal_types:
        return const.COMPILED_TYPE.LITERAL
    # Type
    elif issubclass(schema_type, six.class_types):
        return const.COMPILED_TYPE.TYPE
    # Mapping
    elif isinstance(schema, collections.Mapping):
        return const.COMPILED_TYPE.MAPPING
    # Iterable
    elif isinstance(schema, collections.Iterable):
        return const.COMPILED_TYPE.ITERABLE
    # Callable
    elif callable(schema):
        return const.COMPILED_TYPE.CALLABLE
    # Not detected
    else:
        return None




""" Collection of miscellaneous helpers to alter the validation process. """

import six
import collections
from functools import update_wrapper

from .schema.util import const, get_literal_name
from . import Schema, SchemaError, Invalid
from .validators.boolean import Check


class ObjectProxy(collections.Mapping, dict):
    """ Proxy object attributes as dictionary keys.

    Used with Object() validator
    """

    def __new__(cls, obj):
        """ Factory for object proxies """
        # Use a different class for slots
        if hasattr(obj, '__slots__'):
            cls = SlotsObjectProxy
        # Instantiate
        return super(ObjectProxy, cls).__new__(cls, obj)

    def __init__(self, obj):
        self.obj = obj
        super(ObjectProxy, self).__init__()

    def __len__(self):
        return 0  # we don't care about the length

    def __iter__(self):
        """ Get an iterator for the object attribute names """
        return iter(vars(self.obj))

    def __contains__(self, key):
        return hasattr(self.obj, six.text_type(key))

    def __getitem__(self, key):
        return getattr(self.obj, six.text_type(key))

    def __setitem__(self, key, value):
        return setattr(self.obj, six.text_type(key), value)

    def __delitem__(self, key):
        return delattr(self.obj, six.text_type(key))


class SlotsObjectProxy(ObjectProxy):
    """ Proxy for objects with __slots__ """

    def __iter__(self):
        # Named tuple?
        if isinstance(self.obj, tuple) and hasattr(self.obj, '_fields'):
            return iter(self.obj._fields)
        # Just Slots
        else:
            return iter(self.obj.__slots__)

    def __setitem__(self, key, value):
        try:
            return super(SlotsObjectProxy, self).__setitem__(key, value)
        except AttributeError:
            # Type does not support assignments (e.g. immutable containers)

            # If the value did not change -- okay, let it be
            if value == self[key]:
                return

            # Otherwise -- fail
            raise


class Object(object):
    """ Specify that the provided mapping should validate an object.

    This uses the same mapping validation rules, but works with attributes instead:

    ```python
    from good import Schema, Object

    intify = lambda v: int(v)  # Naive Coerce(int) implementation

    # Define a class to play with
    class Person(object):
        category = u'Something'  # Not validated

        def __init__(self, name, age):
            self.name = name
            self.age = age

    # Schema
    schema = Schema(Object({
        'name': str,
        'age': intify,
    }))

    # Validate
    schema(Person(name=u'Alex', age='18'))  #-> Girl(name=u'Alex', age=18)
    ```

    Internally, it validates the object's `__dict__`: hence, class attributes are excluded from validation.
    Validation is performed with the help of a wrapper class which proxies object attributes as mapping keys,
    and then Schema validates it as a mapping.

    This inherits the default required/extra keys behavior of the Schema.
    To override, use [`Optional()`](#optional) and [`Extra`](#extra) markers.

    :param schema: Object schema, given as a mapping
    :type schema: Mapping
    :param cls: Require instances of a specific class. If `None`, allows all classes.
    :type cls: None|type|tuple[type]
    """

    def __init__(self, schema, cls=None):
        # Input validation
        if not isinstance(schema, collections.Mapping):
            raise SchemaError('Object() argument must be a mapping, {} given'.format(type(schema)))

        # Prepare
        self.name = self._format_cls_name(cls)
        self.cls = object if cls is None else cls

        # Compile schema
        self.compiled = Schema(schema)

    @staticmethod
    def _format_cls_name(c):
        return _(u'Object({cls})').format(cls=c.__name__ if c else u'*')

    def _format_value_type(self, v):
        return self._format_cls_name(type(v)) if isinstance(v, object) else get_literal_name

    def __call__(self, v):
        # Check type
        if not isinstance(v, self.cls):
            raise Invalid(_(u'Wrong value type'), provided=self._format_value_type(v))

        # Validate using ObjectProxy and unwrap
        return self.compiled(ObjectProxy(v)).obj


class Msg(object):
    """ Override the error message reported by the wrapped schema in case of validation errors.

    On validation, if the schema throws [`Invalid`](#invalid) -- the message is overridden with `msg`.

    Some other error types are converted to `Invalid`: see notes on [Schema Callables](#callables).

    ```python
    from good import Schema, Msg

    intify = lambda v: int(v)  # Naive Coerce(int) implementation
    intify.name = u'Number'

    schema = Schema(Msg(intify, u'Need a number'))
    schema(1)  #-> 1
    schema('a')
    #-> Invalid: Need a number: expected Number, got a
    ```

    :param schema: The wrapped schema to modify the error for
    :param message: Error message to use instead of the one that's reported by the underlying schema
    :type message: unicode
    """

    def __init__(self, schema, message):
        assert isinstance(message, six.text_type), 'Msg() message must be a unicode string'
        self.message = message
        self.compiled = Schema(schema).compiled
        self.name = self.compiled.name

    def __getattr__(self, attr):
        """ Inherit all attributes from the wrapped schema """
        return getattr(self.compiled, attr)

    def __call__(self, v):
        try:
            return self.compiled(v)
        except Invalid as ee:
            # Override message
            for e in ee:
                e.message = self.message
            # Raise again
            raise
        except const.transformed_exceptions:
            raise Invalid(self.message or _(u'Invalid value'))



def message(message, name=None):
    """ Convenience decorator that applies [`Msg()`](#msg) to a callable.

    ```python
    from good import Schema, message

    @message(u'Need a number')
    def intify(v):
        return int(v)
    ```

    :param message: Error message to use instead
    :type message: unicode
    :param name: Override schema name as well. See [`name`](#name).
    :type name: None|unicode
    :return: decorator
    :rtype: callable
    """
    def decorator(func):
        wf = update_wrapper(Msg(func, message), func)
        if name:
            wf.name = name
        return wf
    return decorator


def name(name, validator=None):
    """ Set a name on a validator callable.

    Useful for user-friendly reporting when using lambdas to populate the [`Invalid.expected`](#invalid) field:

    ```python
    from good import Schema, name

    Schema(lambda x: int(x))('a')
    #-> Invalid: invalid literal for int(): expected <lambda>(), got
    Schema(name('int()', lambda x: int(x))('a')
    #-> Invalid: invalid literal for int(): expected int(), got a
    ```

    Note that it is only useful with lambdas, since function name is used if available:
    see notes on [Schema Callables](#callables).

    :param name: Name to assign on the validator callable
    :type name: unicode
    :param validator: Validator callable. If not provided -- a decorator is returned instead:

        ```python
        from good import name

        @name(u'int()')
        def int(v):
            return int(v)
        ```

    :type validator: callable
    :return: The same validator callable
    :rtype: callable
    """
    # Decorator mode
    if validator is None:
        def decorator(f):
            f.name = name
            return f
        return decorator

    # Direct mode
    validator.name = name
    return validator


def truth(message, expected=None):
    """ Convenience decorator that applies [`Check`](#check) to a callable.

    ```python
    from good import truth

    @truth(u'Must be an existing directory')
    def isDir(v):
        return os.path.isdir(v)
    ```

    :param message: Validation error message
    :type message: unicode
    :param expected: Expected value string representation, or `None` to get it from the wrapped callable
    :type expected: None|str|unicode
    :return: decorator
    :rtype: callable
    """
    def decorator(func):
        return update_wrapper(Check(func, message, expected), func)
    return decorator

__all__ = ('Object', 'Msg', 'message', 'name', 'truth')



""" A *Marker* is a proxy class which wraps some schema.

Immediately, the example is:

```python
from good import Schema, Required

Schema({
    'name': str,  # required key
    Optional('age'): int,  # optional key
}, default_keys=Required)
```

This way, keys marked with `Required()` will report errors if no value if provided.

Typically, a marker "decorates" a mapping key, but some of them can be "standalone":

```python
from good import Schema, Extra
Schema({
    'name': str,
    Extra: int  # allow any keys, provided their values are integer
})
```

Each marker can have it's own unique behavior since nothing is hardcoded into the core [`Schema`](#schema).
Keep on reading to learn how markers perform.
"""


import six
from .signals import RemoveValue
from .errors import Invalid, MultipleInvalid
from .util import const, get_type_name, get_literal_name


class Marker(object):
    """ A Marker is a class that decorates a mapping key.

    Its compilation goes in 3 phases:

    1. Marker.key is set by the user: `Required('name')`
    2. Marker.key is compiled into a schema, and notified with `Marker.on_compiled(name, key_schema)`
    3. Marker receives a value schema, through `Marker.on_compiled(value_schema=value_schema)`

    Note that if a marker is used as a mapping key, its `key_schema` is compiled as a matcher for performance.

    When CompiledSchema performs matching, it collects input values that match the marker
    (using `priority` to decide which of the markers will actually get the duck)
    and then calls execute() on the Marker so it can implement its logic.

    Note that `execute()` is always called, regardless of whether the marker has matched anything:
    this gives markers a chance to modify the input to its taste.
    This opens the possibilities of implementing custom markers which validate the whole schema!

    Finally, note that a marker does not necessarily decorate something: it can be used as a class:

    ```python
    Schema({
        'name': str,
        Extra: Reject
    })
    ```

    In this case, a Marker class is automatically instantiated
    with an identity function (which matches any value): `Extra(lambda x: x)`.
    """

    #: Marker priority
    #: This defines matching order for mapping keys
    priority = 0

    #: The default marker error message
    #: Stored here just for convenience
    error_message = None

    def __init__(self, key):
        #: The original key
        self.key = key
        #: Human-readable marker representation
        self.name = None
        #: CompiledSchema for the key
        self.key_schema = None
        #: CompiledSchema for value (if the Marker was used as a key in a mapping)
        self.value_schema = None

        #: Whether the marker is used as a mapping key
        self.as_mapping_key = False

    def on_compiled(self, name=None, key_schema=None, value_schema=None, as_mapping_key=None):
        """ When CompiledSchema compiles this marker, it sets informational values onto it.

        Note that arguments may be provided in two incomplete sets,
        e.g. (name, key_schema, None) and then (None, None, value_schema).
        Thus, all assignments must be handled individually.

        It is possible that a marker may have no `value_schema` at all:
        e.g. in the case of { Extra: Reject } -- `Reject` will have no value schema,
        but `Extra` will have compiled `Reject` as the value.

        :param key_schema: Compiled key schema
        :type key_schema: CompiledSchema|None
        :param value_schema: Compiled value schema
        :type value_schema: CompiledSchema|None
        :param name: Human-friendly marker name
        :type name: unicode|None
        :param as_mapping_key: Whether it's used as a mapping key?
        :type as_mapping_key: bool|None
        :rtype: Marker
        """
        if self.name is None:
            self.name = name
        if self.key_schema is None:
            self.key_schema = key_schema
        if self.value_schema is None:
            self.value_schema = value_schema
        if as_mapping_key:
            self.as_mapping_key = True
        return self

    def __repr__(self):
        return '{cls}({0})'.format(
            self.name,
            cls=type(self).__name__)

    #region Marker is a Proxy

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        # Marker equality comparison:
        #  key == key | key == Marker.key | key is Marker
        return self.key == (other.key if isinstance(other, Marker) else other) or other is type(self)

    def __str__(self):
        return six.binary_type(self.key)

    def __unicode__(self):
        return get_literal_name(self.key)

    if six.PY3:
        __bytes__, __str__ = __str__, __unicode__

    #endregion

    def __call__(self, v):
        """ Validate a key using this Marker's schema """
        return self.key_schema(v)

    def execute(self, d, matches):
        """ Execute the marker against the the matching values from the input.

        Note that `execute()` is called precisely once, and even if there are no matches for the marker.

        :param d: The original user input
        :type d: dict
        :param matches: List of (input-key, sanitized-input-key, input-value) triples that matched the given marker
        :type matches: list[tuple]
        :returns: The list of matches, potentially modified
        :rtype: list[tuple]
        :raises: Invalid|MultipleInvalid
        """
        return matches  # No-op by default


class Required(Marker):
    """ `Required(key)` is used to decorate mapping keys and hence specify that these keys must always be present in
    the input mapping.

    When compiled, [`Schema`](#schema) uses `default_keys` as the default marker:

    ```python
    from good import Schema, Required

    schema = Schema({
        'name': str,
        'age': int
    }, default_keys=Required)  # wrap with Required() by default

    schema({'name': 'Mark'})
    #-> Invalid: Required key not provided @ ['age']: expected age, got -none-
    ```

    Remember that mapping keys are schemas as well, and `Require` will expect to always have a match:

    ```python
    schema = Schema({
        Required(str): int,
    })

    schema({})  # no `str` keys provided
    #-> Invalid: Required key not provided: expected String, got -none-
    ```

    In addition, the `Required` marker has special behavior with [`Default`](#default) that allows to set the key
    to a default value if the key was not provided. More details in the docs for [`Default`](#default).
    """
    priority = 0
    error_message = _(u'Required key not provided')

    def execute(self, d, matches):
        # If a Required() key is present -- it expects to ALWAYS have one or more matches

        # When Required() has no matches...
        if not matches:
            # Last chance: value_schema supports Undefined, and the key is a literal
            if self.value_schema.supports_undefined:
                # Schema supports `Undefined`, then use it!
                v = self.value_schema(const.UNDEFINED)
                matches.append((self.key_schema.schema, self.key_schema.schema, v))
                return matches
            else:
                # Invalid
                path = [self.key] if self.key_schema.compiled_type == const.COMPILED_TYPE.LITERAL else []
                raise Invalid(self.error_message, self.name, _(u'-none-'), path)
        return matches


class Optional(Marker):
    """ `Optional(key)` is controversial to [`Required(key)`](#required): specified that the mapping key is not required.

    This only has meaning when a [`Schema`](#schema) has `default_keys=Required`:
    then, it decorates all keys with `Required()`, unless a key is already decorated with some Marker.
    `Optional()` steps in: those keys are already decorated and hence are not wrapped with `Required()`.

    So, it's only used to prevent `Schema` from putting `Required()` on a key.
    In all other senses, it has absolutely no special behavior.

    As a result, optional key can be missing, but if it was provided -- its value must match the value schema.

    Example: use as `default_keys`:

    ```python
    schema = Schema({
        'name': str,
        'age': int
    }, default_keys=Optional)  # Make all keys optional by default

    schema({})  #-> {} -- okay
    schema({'name': None})
    #->  Invalid: Wrong type @ ['name']: expected String, got None
    ```

    Example: use to mark specific keys are not required:

    ```python
    schema = Schema({
        'name': str,
        Optional(str): int  # key is optional
    })

    schema({'name': 'Mark'})  # valid
    schema({'name': 'Mark', 'age': 10})  # valid
    schema({'name': 'Mark', 'age': 'X'})
    #-> Invalid: Wrong type @ ['age']: expected Integer number, got Binary String
    ```
    """

    priority = 0

    pass  # no-op


class Remove(Marker):
    """ `Remove(key)` marker is used to declare that the key, if encountered,
    should be removed, without validating the value.

    `Remove` has highest priority, so it operates before everything else in the schema.

    Example:

    ```python
    schema = Schema({
        Remove('name'): str, # `str` does not mean anything since the key is removed anyway
        'age': int
    })

    schema({'name': 111, 'age': 18})  #-> {'age': 18}
    ```

    However, it's more natural to use `Remove()` on values.
    Remember that in this case `'name'` will become [`Required()`](#required),
    if not decorated with [`Optional()`](#optional):

    ```python
    schema = Schema({
        Optional('name'): Remove
    })

    schema({'name': 111, 'age': 18})  #-> {'age': 18}
    ```

    **Bonus**: `Remove()` can be used in iterables as well:

    ```python
    schema = Schema([str, Remove(int)])
    schema(['a', 'b', 1, 2])  #-> ['a', 'b']
    ```
    """

    priority = 1000  # We always want to remove keys prior to any other actions

    def execute(self, d, matches):
        # Remove all matching keys from the input
        for k, sanitized_k, v in matches:
            d.pop(k)

        # Clean the list of matches so further processing does not assign them again
        return []

    def __call__(self, v):
        if not self.as_mapping_key:
            # When used on a value -- drop it
            raise RemoveValue()
        return super(Remove, self).__call__(v)


class Reject(Marker):
    """ `Reject(key)` marker is used to report [`Invalid`](#invalid) errors every time is matches something in the input.

    It has lower priority than most of other schemas, so rejection will only happen
    if no other schemas has matched this value.

    Example:

    ```python
    schema = Schema({
        Reject('name'): None,  # Reject by key
        Optional('age'): Msg(Reject, u"Field is not supported anymore"), # alternative form
    })

    schema({'name': 111})
    #-> Invalid: Field is not supported anymore @ ['name']: expected -none-, got name
    ```
    """

    priority = -50
    error_message = _(u'Value rejected')

    def __call__(self, v):
        if not self.as_mapping_key:
            # When used on a value -- complain
            raise Invalid(self.error_message, _(u'-none-'), get_literal_name(v), validator=self)
        return super(Reject, self).__call__(v)

    def execute(self, d, matches):
        # Complain on all values it gets
        if matches:
            errors = []
            for k, sanitized_k, v in matches:
                errors.append(Invalid(self.error_message, _(u'-none-'), get_literal_name(k), [k]))
            raise MultipleInvalid.if_multiple(errors)
        return matches


class Allow(Marker):
    """ `Allow(key)` is a no-op marker that never complains on anything.

    Designed to be used with [`Extra`](#extra).
    """
    priority = 0

    pass  # no-op


class Extra(Marker):
    """ `Extra` is a catch-all marker to define the behavior for mapping keys not defined in the schema.

    It has the lowest priority, and delegates its function to its value, which can be a schema, or another marker.

    Given without argument, it's compiled with an identity function `lambda x:x` which is a catch-all:
    it matches any value. Together with lowest priority, `Extra` will only catch values which did not match anything else.

    Every mapping has an `Extra` implicitly, and `extra_keys` argument controls the default behavior.

    Example with `Extra: <schema>`:

    ```python
    schema = Schema({
        'name': str,
        Extra: int  # this will allow extra keys provided they're int
    })

    schema({'name': 'Alex', 'age': 18'})  #-> ok
    schema({'name': 'Alex', 'age': 'X'})
    #-> Invalid: Wrong type @ ['age']: expected Integer number, got Binary String
    ```

    Example with `Extra: Reject`: reject all extra values:

    ```python
    schema = Schema({
        'name': str,
        Extra: Reject
    })

    schema({'name': 'Alex', 'age': 'X'})
    #-> Invalid: Extra keys not allowed @ ['age']: expected -none-, got age
    ```

    Example with `Extra: Remove`: silently discard all extra values:

    ```python
    schema = Schema({'name': str}, extra_keys=Remove)
    schema({'name': 'Alex', 'age': 'X'})  #-> {'name': 'Alex'}
    ```

    Example with `Extra: Allow`: allow any extra values:

    ```python
    schema = Schema({'name': str}, extra_keys=Allow)
    schema({'name': 'Alex', 'age': 'X'})  #-> {'name': 'Alex', 'age': 'X'}
    ```
    """
    priority = -1000  # Extra should match last

    error_message = _(u'Extra keys not allowed')

    def on_compiled(self, name=None, key_schema=None, value_schema=None, as_mapping_key=None):
        # Special case
        # When { Extra: Reject }, use a customized error message
        if value_schema and isinstance(value_schema.compiled, Reject):
            value_schema.compiled.error_message = self.error_message
        return super(Extra, self).on_compiled(name, key_schema, value_schema, as_mapping_key)

    def execute(self, d, matches):
        # Delegate the decision to the value.

        # If the value is a marker -- call execute() on it
        # This is for the cases when `Extra` is mapped to a marker
        if isinstance(self.value_schema.compiled, Marker):
            return self.value_schema.compiled.execute(d, matches)

        # Otherwise, it's a schema, which must be called on every value to validate it.
        # However, CompiledSchema does this anyway at the next step, so doing nothing here
        return matches


class Entire(Optional):
    """ `Entire` is a convenience marker that validates the entire mapping using validators provided as a value.

    It has absolutely lowest priority, lower than `Extra`, hence it never matches any keys, but is still executed to
    validate the mapping itself.

    This opens the possibilities to define rules on multiple fields.
    This feature is leveraged by the [`Inclusive`](#inclusive) and [`Exclusive`](#exclusive) group validators.

    For example, let's require the mapping to have no more than 3 keys:

    ```python
    from good import Schema, Entire

    def maxkeys(n):
        # Return a validator function
        def validator(d):
            # `d` is the dictionary.
            # Validate it
            assert len(d) <= 3, 'Dict size should be <= 3'
            # Return the value since all callable schemas should do that
            return d
        return validator

    schema = Schema({
        str: int,
        Entire: maxkeys(3)
    })
    ```

    In this example, `Entire` is executed for every input dictionary, and magically calls the schema it's mapped to.
    The `maxkeys(n)` schema is a validator that complains on the dictionary size if it's too huge.
    `Schema` catches the `AssertionError` thrown by it and converts it to [`Invalid`](#invalid).

    Note that the schema this marker is mapped to can't replace the mapping object, but it can mutate the given mapping.
    """

    priority = -2000  # Should never match anything

    def execute(self, d, matches):
        # Ignore `matches`, since it's always empty.
        # Instead, pass the mapping `d` to the schema it's mapped to: `value_schema`
        try:
            self.value_schema(d)
        except Invalid as e:
            e.enrich(
                expected=self.value_schema.name,
                provided=get_type_name(type(d)),
                validator=self.value_schema.schema
            )
            raise

        # Still return the same `matches` list
        return matches


__all__ = ('Required', 'Optional', 'Remove', 'Reject', 'Allow', 'Extra', 'Entire')



""" Slim yet handsome validation library.

Core features:

* Simple
* Customizable
* Supports nested model validation
* Error paths (which field contains the error)
* User-friendly error messages
* Internationalization!
* [Robust](misc/performance/performance.md): 10 000 validations per second
* Python 2.7, 3.3+ compatible
* 100% documented and unit-tested

Inspired by the amazing [alecthomas/voluptuous](https://github.com/alecthomas/voluptuous) and 100% compatible with it.
The whole internals have been reworked towards readability and robustness. And yeah, the docs are now exhaustive :)

The rationale for a remake was to make it modular with a tiny core and everything else built on top of that,
ensure that all error messages are user-friendly out of the box, and tweak the performance.
"""

# Init gettext translations, and do not litter the root scope
import gettext as _gettext, six as _six
_gettext.install('good', **({'unicode': True} if _six.PY2 else {}))


# Core

from .schema.errors import SchemaError, Invalid, MultipleInvalid
from .schema.util import register_type_name

from .schema import Schema

from .schema import markers
from .schema.markers import *

# Helpers
from .helpers import *

# Validators
from .validators import *



import os
from ._base import ValidatorBase
from .. import Invalid


class PathExists(ValidatorBase):
    """ Verify that the path exists. """

    name = u'Existing path'

    def __init__(self):
        super(PathExists, self).__init__()

    def __call__(self, v):
        if not os.path.exists(v):
            raise Invalid(_(u'Path does not exist'), provided=u'Missing path')
        return v


class IsFile(PathExists):
    """ Verify that the file exists.

    ```python
    from good import Schema, IsFile

    schema = Schema(IsFile())

    schema('/etc/hosts')  #-> '/etc/hosts'
    schema('/etc')
    #-> Invalid: is not a file: expected Existing file path, got /etc
    ```
    """

    name = u'File path'

    def __call__(self, v):
        super(IsFile, self).__call__(v)
        if not os.path.isfile(v):
            raise Invalid(_(u'Is not a file'), provided=u'Not a file')
        return v


class IsDir(PathExists):
    """ Verify that the directory exists. """

    name = u'Directory path'

    def __call__(self, v):
        super(IsDir, self).__call__(v)
        if not os.path.isdir(v):
            raise Invalid(_(u'Is not a directory'), provided=u'Not a directory')
        return v



__all__ = ('IsFile', 'IsDir', 'PathExists',)
